#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


import os
import argparse


from trezorlib import coins, tx_api
from trezorlib.client import TrezorClient, TrezorClientDebugLink
from trezorlib.transport import get_transport


parser = argparse.ArgumentParser(description="Trezor initializer")

parser.add_argument(
    "--trezor-path", dest="trezor_path", default=None, help="Trezor path"
)
parser.add_argument(
    "--trezor-idx", dest="trezor_idx", default=None, help="Trezor path idx"
)
parser.add_argument(
    "--mnemonic", dest="mnemonic", default=0, type=int, help="Trezor mnemonic"
)
parser.add_argument(
    "--debug",
    dest="debug",
    default=False,
    action="store_const",
    const=True,
    help="Debug",
)
args = parser.parse_args()

mnemonic12 = (
    "alcohol woman abuse must during monitor noble actual mixed trade anger aisle"
)
mnemonic24 = "permit universe parent weapon amused modify essay borrow tobacco budget walnut " "lunch consider gallery ride amazing frog forget treat market chapter velvet useless topple"

debug_mode = args.debug
if args.trezor_path:
    path = args.trezor_path
elif args.trezor_idx:
    path = "bridge:web01" if args.trezor_idx == "usb" else "udp:127.0.0.1:21324"
else:
    path = os.environ.get("TREZOR_PATH", "bridge:web01")

mnemonic = mnemonic24 if args.mnemonic == 1 else mnemonic12


wirelink = get_transport(path)
client = TrezorClientDebugLink(wirelink) if debug_mode else TrezorClient(wirelink)
if debug_mode:
    debuglink = wirelink.find_debug()
    client.set_debuglink(debuglink)

client.transport.session_begin()

client.wipe_device()
client.load_device_by_mnemonic(
    mnemonic=mnemonic,
    pin="",
    passphrase_protection=False,
    label="ph4test",
    language="english",
)

client.transport.session_end()
client.close()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .OntologyOntIdAttribute import OntologyOntIdAttribute

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class OntologyOntIdAddAttributes(p.MessageType):

    def __init__(
        self,
        ont_id: str = None,
        public_key: bytes = None,
        ont_id_attributes: List[OntologyOntIdAttribute] = None,
    ) -> None:
        self.ont_id = ont_id
        self.public_key = public_key
        self.ont_id_attributes = ont_id_attributes if ont_id_attributes is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('ont_id', p.UnicodeType, 0),
            2: ('public_key', p.BytesType, 0),
            3: ('ont_id_attributes', OntologyOntIdAttribute, p.FLAG_REPEATED),
        }



import fileinput
import re
import io
import os
import argparse
import logging
import coloredlogs
import time
import json
import math

from monero_poc.tools.mem_cmp import MemReader


logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.INFO, use_chroot=False)


TICKS_PER_SECOND = 1000000.
MAX_TICKS = 1073741824  # 2**30


def tic2sec(tic):
    return tic/TICKS_PER_SECOND


def remove_ctl(line):
    return re.sub(r'\x1b\[\d+m', '', line.rstrip('\n'))


def avg(iterable):
    return sum(iterable) / float(len(iterable))


def bins(iterable, nbins=1, key=lambda x: x, ceil_bin=False):
    vals = [key(x) for x in iterable]
    min_v = min(vals)
    max_v = max(vals)
    bin_size = ((1 + max_v - min_v) / float(nbins))
    bin_size = math.ceil(bin_size) if ceil_bin else bin_size
    bins = [[] for _ in range(nbins)]
    for c in iterable:
        cv = key(c)
        cbin = int((cv - min_v) / bin_size)
        bins[cbin].append(c)
    return bins


def count_elements(seq) -> dict:
    hist = {}
    for i in seq:
        hist[i] = hist.get(i, 0) + 1
    return hist


def ascii_histogram(seq) -> None:
    """A horizontal frequency-table/histogram plot."""
    counted = count_elements(seq)
    for k in sorted(counted):
        print('{0:5d} {1}'.format(k, '+' * counted[k]))


class LogAnalyzer(object):
    def __init__(self):
        self.args = None
        self.max_line_len = 0
        self.time_prev = 0
        self.time_prev_o = 0
        self.time_ref = 0
        self.time_ref_r = 0
        self.mem_alloc_prev = 0
        self.mem_alloc_ref = None
        self.mem_alloc_max = 0
        self.serial_device = None
        self.tee_file = None
        self.initial_ticks = None
        self.mem_reader = None
        self.mem_diffs_sec = []
        self.mem_diffs_all = []

    def print_line(self, line):
        print(line)
        if self.args.tee_aux:
            self.tee_line(line)

    def tee_line(self, line):
        if self.tee_file:
            self.tee_file.write(line)
            self.tee_file.write('\n')

    def process(self, line):
        line = line.strip()
        if self.args.no_ctl:
            line = remove_ctl(line)
        if not self.args.tee_aux:
            self.tee_line(line)

        m = re.match(r'^(\d+)\s([^\s]+?)\s([^\s]+?)\s(.*)$', line)
        if m is None:
            self.print_line(line)
            return

        ctime_r = time.time()
        ctime = int(m.group(1))
        ctime_o = ctime
        c_prev = self.time_prev
        c_prev_alloc = self.mem_alloc_prev

        if self.args.no_time:
            line = re.sub(r'^\d+\s*', '', line)

        elif self.args.norm_time:
            if self.initial_ticks is None:
                self.initial_ticks = ctime
                ctime = 0
            elif ctime < self.time_prev_o:
                ctime = self.time_prev + (ctime + (1073741824 - self.time_prev_o))
            else:
                ctime = self.time_prev + (ctime - self.time_prev_o)
            line = re.sub(r'^\d+', '%011d' % ctime, line)

        self.time_prev = ctime
        self.time_prev_o = ctime_o

        if '----diagnostic' in line:
            self.time_ref = ctime
            self.time_ref_r = ctime_r
            return

        self.max_line_len = max(self.max_line_len, min(len(line), 140))
        abs_time = tic2sec(ctime - self.time_ref)
        diff_time = tic2sec(ctime - c_prev)

        mem_free = None
        mem_alloc = None

        mmem = re.match(r'.+?F:\s*(\d+)\s*,?\s*A:\s*(\d+)', line)
        if mmem:
            mem_free = int(mmem.group(1))
            mem_alloc = int(mmem.group(2))

        mmem = re.match(r'.+?Free:\s*(\d+)\s*,?\s*Allocated:\s*(\d+)', line)
        if mmem:
            mem_free = int(mmem.group(1))
            mem_alloc = int(mmem.group(2))

        memstr = ''
        is_sec = '====' in line or '####' in line

        if mem_alloc:
            self.mem_alloc_prev = mem_alloc
            self.mem_alloc_max = max(self.mem_alloc_max, mem_alloc)
            if self.mem_alloc_ref is None:
                self.mem_alloc_ref = mem_alloc

            memstr = 'Alloc diff: %6d, refdi: %6d' % (mem_alloc - self.mem_alloc_ref, mem_alloc - c_prev_alloc)

            if self.mem_reader:
                memb = self.mem_reader.next(is_sec)
                if memb:
                    memsuf = ' skipped: %3d' % memb[2] if is_sec and memb[2] > 0 else ''
                    memstr += ' | Alloc cmp diff: %5d %s' % (mem_alloc - memb[0], memsuf)
                    if self.args.full_mem:
                        memline = self.mem_reader.full_line()
                        if self.args.no_ctl:
                            memline = remove_ctl(memline)
                        memstr += ' | ' + memline

                    if is_sec:
                        self.mem_diffs_sec.append(mem_alloc - memb[0])
                    else:
                        self.mem_diffs_all.append(mem_alloc - memb[0])

        if self.args.no_aug:
            self.print_line(line)
            return

        ldiff = self.max_line_len - len(line)
        self.print_line('%s%s |  AbsTime: %7.3f,   Diff %5.3f  | %s' % (line, ' '*ldiff, abs_time, diff_time, memstr))

        if is_sec:
            abs_r = ''
            if self.serial_device:
                abs_time_r = ctime_r - self.time_ref_r
                abs_r = 'r: %7.2f, ticks p.s.: %7.3f' % (abs_time_r, (ctime - self.time_ref) / float(abs_time_r))
            self.print_line(' ++ TOTAL: %7.2f, %s mem max: %s' % (abs_time, abs_r, self.mem_alloc_max - self.mem_alloc_ref if self.mem_alloc_ref is not None else '?'))

            self.time_ref = ctime
            self.time_ref_r = ctime_r
            if mem_alloc is not None:
                self.mem_alloc_ref = mem_alloc
                self.mem_alloc_max = mem_alloc

    def read_serial(self, device, brate):
        try:
            import serial  # pip install pyserial
        except ImportError:
            raise ValueError('pip install pyserial')

        while True:
            try:
                ser = serial.Serial(device, brate, timeout=.1)
                logger.info('Connected: %s' % ser)

                sio = io.TextIOWrapper(io.BufferedRWPair(ser, ser))
                while True:
                    line = sio.readline()
                    line = line.strip()
                    if len(line) == 0:
                        continue

                    self.process(line)

            except Exception as e:
                logger.warning('Exc: %s' % e)
                time.sleep(2)

    def read_files(self, files):
        for idx, line in enumerate(fileinput.input(files)):
            anz.process(line)

        if self.mem_diffs_sec:
            print('Mem diff avg, sec: %8.3f, all: %8.3f' % (avg(self.mem_diffs_sec), avg(self.mem_diffs_all)))

            if self.args.mem_json:
                print('JSON sec difs: ')
                print(json.dumps(self.mem_diffs_sec))
                print('JSON all difs: ')
                print(json.dumps(self.mem_diffs_all))

            if self.args.mem_bin:
                print('Binned memory differences:')
                for bi in bins(self.mem_diffs_all, 100):
                    if len(bi) == 0:
                        continue
                    print('[%5d, %5d] (avg = %5d): %5d' % (min(bi), max(bi), avg(bi), len(bi)))

    def main(self):
        parser = argparse.ArgumentParser(description='Trezor log reader and parser')
        parser.add_argument('--serial', default=None,
                            help='Serial device to read from')
        parser.add_argument('--brate', type=int, default=115200,
                            help='Baud rate')
        parser.add_argument("--retry", dest="retry", default=True, action="store_const", const=True,
                            help="Retry reconnect")
        parser.add_argument("--norm-time", dest="norm_time", default=False, action="store_const", const=True,
                            help="Normalize time")
        parser.add_argument("--no-time", dest="no_time", default=False, action="store_const", const=True,
                            help="Do not show time")
        parser.add_argument("--no-aug", dest="no_aug", default=False, action="store_const", const=True,
                            help="Do not augment the log output")
        parser.add_argument("--tee", dest="tee", default=None,
                            help="File to copy raw output to")
        parser.add_argument("--tee-aux", dest="tee_aux", default=False, action="store_const", const=True,
                            help="Tee augmented lines")
        parser.add_argument("--tee-append", dest="tee_append", default=False, action="store_const", const=True,
                            help="Append to the tee file")
        parser.add_argument("--no-ctl", dest="no_ctl", default=False, action="store_const", const=True,
                            help="Removes shell controll sequences")
        parser.add_argument("--mem-file", dest="mem_file", default=None,
                            help="Memory log to compare to")
        parser.add_argument("--full-mem", dest="full_mem", default=False, action="store_const", const=True,
                            help="Prints full corresponding memory line")
        parser.add_argument("--mem-json", dest="mem_json", default=False, action="store_const", const=True,
                            help="Dumps mem differences in json")
        parser.add_argument("--mem-bin", dest="mem_bin", default=False, action="store_const", const=True,
                            help="Dumps mem differences binned")
        parser.add_argument('files', metavar='FILE', nargs='*',
                            help='files to read, if empty, stdin is used')
        args = parser.parse_args()

        self.args = args
        if args.tee:
            ex = os.path.exists(args.tee)
            if ex and not args.tee_append:
                raise ValueError('Tee file already exists')

            self.tee_file = open(args.tee, 'w+' if not ex else 'a+')
            if ex:
                self.tee_file.write(('='*80) + (' Time: %s' % time.time()))

        if args.mem_file:
            with open(args.mem_file, 'r') as fh:
                self.mem_reader = MemReader(fh.readlines())

        try:
            if args.serial:
                self.serial_device = args.serial
                self.read_serial(args.serial, args.brate)
            else:
                self.read_files(args.files)

        except KeyboardInterrupt:
            logger.info('Terminating')

        if self.tee_file:
            self.tee_file.close()


if __name__ == '__main__':
    anz = LogAnalyzer()
    anz.main()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class SignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 15

    def __init__(
        self,
        outputs_count: int = None,
        inputs_count: int = None,
        coin_name: str = None,
        version: int = None,
        lock_time: int = None,
        expiry: int = None,
        overwintered: bool = None,
    ) -> None:
        self.outputs_count = outputs_count
        self.inputs_count = inputs_count
        self.coin_name = coin_name
        self.version = version
        self.lock_time = lock_time
        self.expiry = expiry
        self.overwintered = overwintered

    @classmethod
    def get_fields(cls):
        return {
            1: ('outputs_count', p.UVarintType, 0),  # required
            2: ('inputs_count', p.UVarintType, 0),  # required
            3: ('coin_name', p.UnicodeType, 0),  # default=Bitcoin
            4: ('version', p.UVarintType, 0),  # default=1
            5: ('lock_time', p.UVarintType, 0),  # default=0
            6: ('expiry', p.UVarintType, 0),
            7: ('overwintered', p.BoolType, 0),
        }



import fileinput
import re
import argparse
import logging
import coloredlogs

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.INFO, use_chroot=False)


def remove_ctl(line):
    return re.sub(r'\x1b\[\d+m', '', line.rstrip('\n'))


class CtlRemover(object):
    def __init__(self):
        pass

    def process(self, line):
        return remove_ctl(line)

    def read_files(self, files):
        for idx, line in enumerate(fileinput.input(files)):
            print(anz.process(line))

    def main(self):
        parser = argparse.ArgumentParser(description='Removes shell controll sequences (e.g., colouring)')
        parser.add_argument('files', metavar='FILE', nargs='*', help='files to read, if empty, stdin is used')
        args = parser.parse_args()

        try:
            self.read_files(args.files)

        except KeyboardInterrupt:
            logger.info('Terminating')


if __name__ == '__main__':
    anz = CtlRemover()
    anz.main()



from monero_glue.xmr import crypto
from monero_glue.xmr.sub.addr import encode_addr
from monero_glue.xmr.sub.xmr_net import NetworkTypes, net_version


class AccountCreds(object):
    """
    Stores account private keys
    """

    def __init__(
        self,
        view_key_private=None,
        spend_key_private=None,
        view_key_public=None,
        spend_key_public=None,
        address=None,
        network_type=NetworkTypes.MAINNET,
    ):
        self.view_key_private = view_key_private
        self.view_key_public = view_key_public
        self.spend_key_private = spend_key_private
        self.spend_key_public = spend_key_public
        self.address = address
        self.network_type = network_type
        self.multisig_keys = []

    @classmethod
    def new_wallet(
        cls, priv_view_key, priv_spend_key, network_type=NetworkTypes.MAINNET
    ):
        pub_view_key = crypto.scalarmult_base(priv_view_key)
        pub_spend_key = crypto.scalarmult_base(priv_spend_key)
        addr = encode_addr(
            net_version(network_type),
            crypto.encodepoint(pub_spend_key),
            crypto.encodepoint(pub_view_key),
        )
        return cls(
            view_key_private=priv_view_key,
            spend_key_private=priv_spend_key,
            view_key_public=pub_view_key,
            spend_key_public=pub_spend_key,
            address=addr,
            network_type=network_type,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
from binascii import unhexlify
import unittest

import aiounittest
from monero_glue.xmr import common, crypto
from monero_glue.xmr.core import ec_py


class CryptoTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(CryptoTest, self).__init__(*args, **kwargs)

    def test_ed_crypto(self):
        sqr = ec_py.fe_expmod(ec_py.py_fe_sqrtm1, 2)
        self.assertEqual(sqr, ec_py.fe_mod(-1))
        self.assertEqual(
            ec_py.py_fe_A, ec_py.fe_mod(2 * (1 - ec_py.d) * ec_py.inv(1 + ec_py.py_d))
        )

        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb1, 2),
            ec_py.fe_mod(-2 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb2, 2),
            ec_py.fe_mod(2 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb3, 2),
            ec_py.fe_mod(-ec_py.py_fe_sqrtm1 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )
        self.assertEqual(
            ec_py.fe_expmod(ec_py.py_fe_fffb4, 2),
            ec_py.fe_mod(ec_py.py_fe_sqrtm1 * ec_py.py_fe_A * (ec_py.py_fe_A + 2)),
        )

    def test_encoding(self):
        point = unhexlify(
            b"2486224797d05cae3cba4be043be2db0df381f3f19cfa113f86ab38e3d8d2bd0"
        )
        self.assertEqual(point, crypto.encodepoint(crypto.decodepoint(point)))
        self.assertTrue(
            crypto.point_eq(
                crypto.decodepoint(point),
                crypto.decodepoint(crypto.encodepoint(crypto.decodepoint(point))),
            )
        )

    def test_scalarmult_base(self):
        scalar = crypto.decodeint(
            unhexlify(
                b"a0eea49140a3b036da30eacf64bd9d56ce3ef68ba82ef13571ec511edbcf8303"
            )
        )
        exp = unhexlify(
            b"16bb4a3c44e2ced511fc0d4cd86b13b3af21efc99fb0356199fac489f2544c09"
        )
        res = crypto.scalarmult_base(scalar)
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

        scalar = crypto.decodeint(
            unhexlify(
                b"fd290dce39f781aebbdbd24584ed6d48bd300de19d9c3decfda0a6e2c6751d0f"
            )
        )
        exp = unhexlify(
            b"123daf90fc26f13c6529e6b49bfed498995ac383ef19c0db6771143f24ba8dd5"
        )
        res = crypto.scalarmult_base(scalar)
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

    def test_scalarmult(self):
        priv = unhexlify(
            b"3482fb9735ef879fcae5ec7721b5d3646e155c4fb58d6cc11c732c9c9b76620a"
        )
        pub = unhexlify(
            b"2486224797d05cae3cba4be043be2db0df381f3f19cfa113f86ab38e3d8d2bd0"
        )
        exp = unhexlify(
            b"adcd1f5881f46f254900a03c654e71950a88a0236fa0a3a946c9b8daed6ef43d"
        )
        res = crypto.scalarmult(crypto.decodepoint(pub), crypto.decodeint(priv))
        self.assertEqual(exp, crypto.encodepoint(res))
        self.assertTrue(crypto.point_eq(crypto.decodepoint(exp), res))

    def test_cn_fast_hash(self):
        inp = unhexlify(
            b"259ef2aba8feb473cf39058a0fe30b9ff6d245b42b6826687ebd6b63128aff6405"
        )
        res = crypto.cn_fast_hash(inp)
        self.assertEqual(
            res,
            unhexlify(
                b"86db87b83fb1246efca5f3b0db09ce3fa4d605b0d10e6507cac253dd31a3ec16"
            ),
        )

    def test_hash_to_scalar(self):
        inp = unhexlify(
            b"259ef2aba8feb473cf39058a0fe30b9ff6d245b42b6826687ebd6b63128aff6405"
        )
        res = crypto.hash_to_scalar(inp)
        exp = crypto.decodeint(binascii.unhexlify(
            b"9907925b254e12162609fc0dfd0fef2aa4d605b0d10e6507cac253dd31a3ec06"))
        self.assertTrue(crypto.sc_eq(res, exp))

    def test_hash_to_point(self):
        data = unhexlify(
            b"42f6835bf83114a1f5f6076fe79bdfa0bd67c74b88f127d54572d3910dd09201"
        )
        res = crypto.hash_to_point(data)
        res_p = crypto.encodepoint(res)
        self.assertEqual(
            res_p,
            unhexlify(
                b"54863a0464c008acc99cffb179bc6cf34eb1bbdf6c29f7a070a7c6376ae30ab5"
            ),
        )

    def test_derivation_to_scalar(self):
        derivation = unhexlify(
            b"e720a09f2e3a0bbf4e4ba7ad93653bb296885510121f806acb2a5f9168fafa01"
        )
        scalar = unhexlify(
            b"25d08763414c379aa9cf989cdcb3cadd36bd5193b500107d6bf5f921f18e470e"
        )
        sc_int = crypto.derivation_to_scalar(crypto.decodepoint(derivation), 0)
        self.assertEqual(scalar, crypto.encodeint(sc_int))

    def test_generate_key_derivation(self):
        key_pub = crypto.decodepoint(
            unhexlify(
                b"7739c95d3298e2f87362dba9e0e0b3980a692ae8e2f16796b0e382098cd6bd83"
            )
        )
        key_priv = crypto.decodeint(
            unhexlify(
                b"3482fb9735ef879fcae5ec7721b5d3646e155c4fb58d6cc11c732c9c9b76620a"
            )
        )
        deriv_exp = unhexlify(
            b"fa188a45a0e4daccc0e6d4f6f6858fd46392104be74183ec0047e7e9f4eaf739"
        )
        self.assertEqual(
            deriv_exp,
            crypto.encodepoint(crypto.generate_key_derivation(key_pub, key_priv)),
        )

    def test_h(self):
        H = unhexlify(
            b"8b655970153799af2aeadc9ff1add0ea6c7251d54154cfa92c173a0dd39c1f94"
        )
        self.assertEqual(crypto.encodepoint(crypto.xmr_H()), H)

    def test_h_pow(self):
        hp = crypto.gen_Hpow(10)
        self.assertEqual(crypto.encodepoint(hp[0]), crypto.encodepoint(crypto.xmr_H()))
        for i in range(1, 10):
            crypto.check_ed25519point(hp[i])
            self.assertEqual(
                crypto.encodepoint(hp[i]),
                crypto.encodepoint(
                    crypto.scalarmult(crypto.xmr_H(), crypto.sc_init(2 ** i))
                ),
            )

    def test_signature(self):
        for i in range(10):
            priv = crypto.random_scalar()
            data = crypto.cn_fast_hash(bytes(bytearray([i])))

            c, r, pub = crypto.generate_signature(data, priv)
            res = crypto.check_signature(data, c, r, pub)
            self.assertEqual(res, 1)

            res2 = crypto.check_signature(
                data, crypto.sc_add(c, crypto.sc_init(1)), r, pub
            )
            self.assertEqual(res2, 0)

    def test_edhex(self):
        inputs = [crypto.q - 2 ** 9, crypto.q - 10, 0, 100, 2 ** 200 + 10] + [
            common.rand.randrange(0, crypto.q - 2) for _ in range(20)
        ]

        for x in inputs:
            l = crypto.encode_ed25519(x)
            d = crypto.decode_ed25519(l)
            self.assertEqual(x, d)

    def test_modm(self):
        inputs = [crypto.l - 2 ** 9, crypto.l - 10, 0, 100, 2 ** 200 + 10] + [
            common.rand.randrange(0, crypto.l - 2) for _ in range(20)
        ]

        for x in inputs:
            l = crypto.encode_modm(x)
            d = crypto.decode_modm(l)
            self.assertEqual(x, d)

    def test_ge25519_double_scalarmult_vartime2(self):
        for i in range(10):
            ap = crypto.random_scalar()
            bp = crypto.random_scalar()
            A = crypto.scalarmult_base(ap)
            B = crypto.scalarmult_base(bp)
            a = crypto.random_scalar()
            b = crypto.random_scalar()

            R = crypto.ge_double_scalarmult_base_vartime2(a, A, b, B)
            R_exp = crypto.point_add(crypto.scalarmult(A, a), crypto.scalarmult(B, b))
            self.assertTrue(crypto.point_eq(R, R_exp))

    def test_ge25519_double_scalarmult_vartime(self):
        for i in range(10):
            ap = crypto.random_scalar()
            A = crypto.scalarmult_base(ap)
            a = crypto.random_scalar()
            b = crypto.random_scalar()

            R = crypto.ge_double_scalarmult_base_vartime(a, A, b)
            R_exp = crypto.point_add(crypto.scalarmult(A, a), crypto.scalarmult_base(b))
            self.assertTrue(crypto.point_eq(R, R_exp))

    def test_pointadd(self):
        a = crypto.random_scalar()
        A = crypto.scalarmult_base(a)
        A2 = crypto.point_add(A, A)
        A3 = crypto.point_add(A2, A)
        A4 = crypto.point_add(A3, A)
        A8 = crypto.scalarmult(A4, crypto.sc_init(2))

        A8p = crypto.point_mul8(A)
        self.assertTrue(crypto.point_eq(A8p, A8))
        self.assertTrue(crypto.point_eq(A4, crypto.scalarmult(A, crypto.sc_init(4))))
        self.assertTrue(crypto.point_eq(A3, crypto.scalarmult(A, crypto.sc_init(3))))

    def test_sc_inversion(self):
        res = crypto.new_scalar()
        inp = crypto.decodeint(
            unhexlify(
                b"3482fb9735ef879fcae5ec7721b5d3646e155c4fb58d6cc11c732c9c9b76620a"
            )
        )

        crypto.sc_inv_into(res, inp)
        self.assertEqual(
            binascii.hexlify(crypto.encodeint(res)),
            b"bcf365a551e6358f3f281a6241d4a25eded60230b60a1d48c67b51a85e33d70e",
        )
        

if __name__ == "__main__":
    unittest.main()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import logging
import traceback

from monero_glue.agent import agent_misc
from monero_glue.hwtoken import misc as tmisc
from monero_glue.messages import (
    Failure,
    MoneroGetAddress,
    MoneroGetWatchKey,
    MoneroKeyImageSyncFinalRequest,
    MoneroKeyImageSyncStepRequest,
    MoneroTransactionAllInputsSetAck,
    MoneroTransactionAllInputsSetRequest,
    MoneroTransactionAllOutSetAck,
    MoneroTransactionAllOutSetRequest,
    MoneroTransactionData,
    MoneroTransactionFinalAck,
    MoneroTransactionFinalRequest,
    MoneroTransactionInitAck,
    MoneroTransactionInitRequest,
    MoneroTransactionInputsPermutationRequest,
    MoneroTransactionInputViniRequest,
    MoneroTransactionMlsagDoneAck,
    MoneroTransactionMlsagDoneRequest,
    MoneroTransactionRsigData,
    MoneroTransactionSetInputAck,
    MoneroTransactionSetInputRequest,
    MoneroTransactionSetOutputAck,
    MoneroTransactionSetOutputRequest,
    MoneroTransactionSignInputAck,
    MoneroTransactionSignInputRequest,
)
from monero_glue.protocol_base.base import TError
from monero_glue.xmr import common, crypto, key_image, monero, ring_ct
from monero_glue.xmr.enc import chacha_poly
from monero_glue.xmr.sub import addr as xmr_addr
from monero_serialize import xmrserialize, xmrtypes

logger = logging.getLogger(__name__)


DEFAULT_MONERO_BIP44 = [
    0x8000002c,
    0x80000080,
    0x80000000,
    0,
    0,
]  # parse_path(monero.DEFAULT_BIP32_PATH)


def get_rsig_type(us_bulletproof, num_outputs):
    if not us_bulletproof:
        return 0  # Borromean
    elif num_outputs > 16:
        return 2  # Multioutputs
    else:
        return 3  # Padded


def generate_rsig_batch_sizes(rsig_type, num_outputs):
    amount_batched = 0
    batches = []

    while amount_batched < num_outputs:
        if rsig_type == 0 or rsig_type == 1:  # Borromean, BP per output
            batches.append(1)
            amount_batched += 1

        elif rsig_type == 3:  # BP padded
            if num_outputs > 16:
                raise ValueError(
                    "BP padded can support only BULLETPROOF_MAX_OUTPUTS statements"
                )

            batches.append(num_outputs)
            amount_batched += num_outputs

        elif rsig_type == 2:  # multioutput
            batch_size = 1
            while (
                batch_size * 2 + amount_batched <= num_outputs and batch_size * 2 <= 16
            ):
                batch_size *= 2
            batch_size = min(batch_size, num_outputs - amount_batched)
            batches.append(batch_size)
            amount_batched += batch_size

        else:
            raise ValueError("Unknown rsig type")
    return batches


class TData(object):
    """
    Agent transaction-scoped data
    """

    def __init__(self):
        self.tsx_data = None  # type: MoneroTransactionData
        self.tx_data = None  # construction data
        self.rsig_type = 0
        self.rsig_batches = []
        self.rsig_param = None
        self.cur_input_idx = 0
        self.cur_output_idx = 0
        self.cur_batch_idx = 0
        self.cur_output_in_batch_idx = 0
        self.tx = xmrtypes.Transaction(version=2, vin=[], vout=[], extra=[])
        self.tx_in_hmacs = []
        self.tx_out_entr_hmacs = []
        self.tx_out_hmacs = []
        self.tx_out_rsigs = []
        self.tx_out_pk = []
        self.tx_out_ecdh = []
        self.source_permutation = []
        self.alphas = []
        self.spend_encs = []
        self.pseudo_outs = []
        self.couts = []
        self.rsig_gamma = []
        self.tx_prefix_hash = None
        self.enc_salt1 = None
        self.enc_salt2 = None
        self.enc_keys = None  # encrypted tx keys
        self.rv = None


class SignAuxData(object):
    def __init__(self):
        self.destinations = None  # type: list[tuple[xmr_addr.AddrInfo, int]]


class Agent(object):
    """
    Glue agent, running on host
    """

    def __init__(
        self, trezor, address_n=None, network_type=None, slip0010=False, **kwargs
    ):
        self.trezor = trezor
        self.ct = None  # type: TData
        self.address_n = address_n if address_n else DEFAULT_MONERO_BIP44
        self.network_type = network_type
        if slip0010 and 0 in address_n:
            raise ValueError("SLIP0010 cannto contain public derivation")
        if slip0010 and address_n is None:
            self.address_n = DEFAULT_MONERO_BIP44[:-2]

    def is_simple(self, rv):
        """
        True if simpe
        :param rv:
        :return:
        """
        return rv.type in [xmrtypes.RctType.Simple, xmrtypes.RctType.FullBulletproof]

    def is_bulletproof(self, rv):
        """
        True if bulletproof
        :param rv:
        :return:
        """
        return rv.type in [
            xmrtypes.RctType.FullBulletproof,
            xmrtypes.RctType.SimpleBulletproof,
        ]

    def is_error(self, response):
        """
        True if trezor returned an error
        :param response:
        :return:
        """
        return isinstance(response, TError) or isinstance(response, Failure)

    def handle_error(self, response):
        """
        Raises an error if Trezor returned an error.
        :param response:
        :return:
        """
        if not self.is_error(response):
            return
        raise agent_misc.TrezorReturnedError(resp=response)

    async def sign_unsigned_tx(self, unsig):
        """
        Processes unsigned transaction set, returns serialized signed transactions.
        :param unsig:
        :return:
        """
        txes = []
        for tx in unsig.txes:
            await self.sign_transaction_data(tx)
            txes.append(await self.serialized_tx())

        return txes

    async def sign_tx(self, construction_data):
        """
        Transfers transaction, serializes response

        :param construction_data:
        :return:
        """
        if not isinstance(construction_data, list):
            construction_data = [construction_data]

        txes = []
        for tdata in construction_data:
            await self.sign_transaction_data(tdata)
            txes.append(await self.serialized_tx())

        return txes

    async def serialized_tx(self):
        """
        Returns the last signed transaction as blob
        :return:
        """
        return await self.serialize_tx(self.ct.tx)

    async def serialize_tx(self, tx):
        """
        Serializes transaction
        :param tx:
        :return:
        """
        writer = xmrserialize.MemoryReaderWriter()
        ar1 = xmrserialize.Archive(writer, True)
        await ar1.message(tx, msg_type=xmrtypes.Transaction)
        return bytes(writer.get_buffer())

    def _is_req_bulletproof(self):
        return self.ct.rsig_type != 0

    def _is_bulletproof(self):
        if self.ct.rv is None:
            raise ValueError("RV is None")
        return self.ct.rv.type == 3

    def _is_simple(self):
        if self.ct.rv is None:
            raise ValueError("RV is None")
        return self.ct.rv.type == 2

    def _is_offloading(self):
        return self.ct and self.ct.rsig_param and self.ct.rsig_param.offload_type != 0

    def _num_outputs(self):
        return len(self.ct.tx_data.splitted_dsts)

    def _num_inputs(self):
        return len(self.ct.tx_data.sources)

    def _get_integrated_idx(self, outputs, aux_data=None):
        if aux_data is None or aux_data.destinations is None:
            return []

        integrated_indices = []
        integrated_pairs = []
        for idx, cur in enumerate(aux_data.destinations):
            ainfo = cur[0]
            if ainfo.is_integrated:
                caddr = xmr_addr.build_address(ainfo.spend_key, ainfo.view_key)
                integrated_pairs.append((caddr, cur[1]))

        for idx, dst in enumerate(outputs):
            for ipair in integrated_pairs:
                if xmr_addr.addr_eq(ipair[0], dst.addr) and dst.amount == ipair[1]:
                    integrated_indices.append(idx)

        logger.debug("Integrated indices: %s" % len(integrated_indices))
        return integrated_indices

    async def sign_transaction_data(
        self,
        tx,
        multisig=False,
        exp_tx_prefix_hash=None,
        use_tx_keys=None,
        aux_data=None,
    ):
        """
        Uses Trezor to sign the transaction
        :param tx:
        :type tx: xmrtypes.TxConstructionData
        :param multisig:
        :param exp_tx_prefix_hash:
        :param use_tx_keys:
        :param aux_data:
        :return:
        """
        self.ct = TData()
        self.ct.tx_data = tx

        payment_id = []
        extras = await monero.parse_extra_fields(tx.extra)
        extra_nonce = monero.find_tx_extra_field_by_type(extras, xmrtypes.TxExtraNonce)
        if extra_nonce and monero.has_encrypted_payment_id(extra_nonce.nonce):
            payment_id = monero.get_encrypted_payment_id_from_tx_extra_nonce(
                extra_nonce.nonce
            )
        elif extra_nonce and monero.has_payment_id(extra_nonce.nonce):
            payment_id = monero.get_payment_id_from_tx_extra_nonce(extra_nonce.nonce)

        # Init transaction
        tsx_data = MoneroTransactionData()
        tsx_data.version = 1
        tsx_data.payment_id = payment_id
        tsx_data.unlock_time = tx.unlock_time
        tsx_data.outputs = [
            tmisc.translate_monero_dest_entry_pb(x) for x in tx.splitted_dsts
        ]
        tsx_data.change_dts = tmisc.translate_monero_dest_entry_pb(tx.change_dts)
        tsx_data.num_inputs = len(tx.sources)
        tsx_data.mixin = len(tx.sources[0].outputs)
        tsx_data.fee = sum([x.amount for x in tx.sources]) - sum(
            [x.amount for x in tx.splitted_dsts]
        )
        tsx_data.account = tx.subaddr_account
        tsx_data.minor_indices = tx.subaddr_indices
        tsx_data.is_multisig = multisig
        tsx_data.exp_tx_prefix_hash = common.defval(exp_tx_prefix_hash, b"")
        tsx_data.use_tx_keys = common.defval(use_tx_keys, [])
        tsx_data.integrated_indices = self._get_integrated_idx(
            tsx_data.outputs, aux_data
        )
        self.ct.tx.unlock_time = tx.unlock_time

        # Rsig
        num_outputs = len(tsx_data.outputs)
        use_bp = common.defattr(tx, "use_bulletproofs", False)
        self.ct.rsig_type = get_rsig_type(use_bp, num_outputs)
        self.ct.rsig_batches = generate_rsig_batch_sizes(self.ct.rsig_type, num_outputs)
        rsig_data = MoneroTransactionRsigData(
            rsig_type=self.ct.rsig_type, grouping=self.ct.rsig_batches
        )
        tsx_data.rsig_data = rsig_data

        self.ct.tsx_data = tsx_data
        init_msg = MoneroTransactionInitRequest(
            version=0,
            address_n=self.address_n,
            network_type=self.network_type,
            tsx_data=tsx_data,
        )

        t_res = await self.trezor.tsx_sign(init_msg)  # type: MoneroTransactionInitAck
        self.handle_error(t_res)

        in_memory = t_res.in_memory
        self.ct.tx_out_entr_hmacs = t_res.hmacs
        self.ct.rsig_param = t_res.rsig_data

        # Set transaction inputs
        for idx, src in enumerate(tx.sources):
            src_bin = tmisc.translate_monero_src_entry_pb(src)
            msg = MoneroTransactionSetInputRequest(src_entr=src_bin)

            t_res = await self.trezor.tsx_sign(
                msg
            )  # type: MoneroTransactionSetInputAck
            self.handle_error(t_res)

            vini_p = await tmisc.parse_msg(t_res.vini, xmrtypes.TxInV())
            vini = vini_p.txin_to_key
            self.ct.tx.vin.append(vini)
            self.ct.tx_in_hmacs.append(t_res.vini_hmac)
            self.ct.pseudo_outs.append((t_res.pseudo_out, t_res.pseudo_out_hmac))
            self.ct.alphas.append(t_res.alpha_enc)
            self.ct.spend_encs.append(t_res.spend_enc)

        # Sort key image
        self.ct.source_permutation = list(range(len(tx.sources)))
        self.ct.source_permutation.sort(
            key=lambda x: self.ct.tx.vin[x].k_image, reverse=True
        )

        def swapper(x, y):
            self.ct.tx.vin[x], self.ct.tx.vin[y] = self.ct.tx.vin[y], self.ct.tx.vin[x]
            self.ct.tx_in_hmacs[x], self.ct.tx_in_hmacs[y] = (
                self.ct.tx_in_hmacs[y],
                self.ct.tx_in_hmacs[x],
            )
            self.ct.pseudo_outs[x], self.ct.pseudo_outs[y] = (
                self.ct.pseudo_outs[y],
                self.ct.pseudo_outs[x],
            )
            self.ct.alphas[x], self.ct.alphas[y] = self.ct.alphas[y], self.ct.alphas[x]
            self.ct.spend_encs[x], self.ct.spend_encs[y] = (
                self.ct.spend_encs[y],
                self.ct.spend_encs[x],
            )
            tx.sources[x], tx.sources[y] = tx.sources[y], tx.sources[x]

        common.apply_permutation(self.ct.source_permutation, swapper)

        if not in_memory:
            msg = MoneroTransactionInputsPermutationRequest(
                perm=self.ct.source_permutation
            )
            t_res = await self.trezor.tsx_sign(msg)
            self.handle_error(t_res)

        # Set vin_i back - tx prefix hashing
        # Done only if not in-memory.
        if not in_memory:
            for idx in range(len(self.ct.tx.vin)):
                msg = MoneroTransactionInputViniRequest(
                    src_entr=tmisc.translate_monero_src_entry_pb(tx.sources[idx]),
                    vini=await tmisc.dump_msg(self.ct.tx.vin[idx], prefix=b"\x02"),
                    vini_hmac=self.ct.tx_in_hmacs[idx],
                    pseudo_out=self.ct.pseudo_outs[idx][0] if not in_memory else None,
                    pseudo_out_hmac=self.ct.pseudo_outs[idx][1]
                    if not in_memory
                    else None,
                )
                t_res = await self.trezor.tsx_sign(msg)
                self.handle_error(t_res)

        # All inputs set
        t_res = await self.trezor.tsx_sign(
            MoneroTransactionAllInputsSetRequest()
        )  # type: MoneroTransactionAllInputsSetAck
        self.handle_error(t_res)
        await self._on_all_input_set(t_res)

        # Set transaction outputs
        for idx, dst in enumerate(tx.splitted_dsts):
            msg = await self._step_set_outputs(idx, dst)

            t_res = await self.trezor.tsx_sign(
                msg
            )  # type: MoneroTransactionSetOutputAck
            self.handle_error(t_res)

            await self._on_set_outputs_ack(t_res)

        t_res = await self.trezor.tsx_sign(
            MoneroTransactionAllOutSetRequest()
        )  # type: MoneroTransactionAllOutSetAck
        self.handle_error(t_res)

        rv = xmrtypes.RctSig()
        rv.p = xmrtypes.RctSigPrunable()
        rv.txnFee = t_res.rv.txn_fee
        rv.message = t_res.rv.message
        rv.type = t_res.rv.rv_type
        self.ct.tx.extra = list(bytearray(t_res.extra))

        # Verify transaction prefix hash correctness, tx hash in one pass
        self.ct.tx_prefix_hash = await monero.get_transaction_prefix_hash(self.ct.tx)
        if not common.ct_equal(t_res.tx_prefix_hash, self.ct.tx_prefix_hash):
            raise ValueError("Transaction prefix has does not match")

        # RctSig
        if self.is_simple(rv):
            if self.is_bulletproof(rv):
                rv.p.pseudoOuts = [x[0] for x in self.ct.pseudo_outs]
            else:
                rv.pseudoOuts = [x[0] for x in self.ct.pseudo_outs]

        # Range proof
        rv.p.rangeSigs = []
        rv.p.bulletproofs = []
        rv.outPk = []
        rv.ecdhInfo = []
        for idx in range(len(self.ct.tx_out_pk)):
            rv.outPk.append(self.ct.tx_out_pk[idx])
            rv.ecdhInfo.append(self.ct.tx_out_ecdh[idx])

        for idx in range(len(self.ct.tx_out_rsigs)):
            if self.is_bulletproof(rv):
                rv.p.bulletproofs.append(self.ct.tx_out_rsigs[idx])
            else:
                rv.p.rangeSigs.append(self.ct.tx_out_rsigs[idx])

        # MLSAG message check
        t_res = await self.trezor.tsx_sign(
            MoneroTransactionMlsagDoneRequest()
        )  # type: MoneroTransactionMlsagDoneAck
        self.handle_error(t_res)

        mlsag_hash = t_res.full_message_hash
        mlsag_hash_computed = await monero.get_pre_mlsag_hash(rv)
        if not common.ct_equal(mlsag_hash, mlsag_hash_computed):
            raise ValueError("Pre MLSAG hash has does not match")

        # Sign each input
        couts = []
        rv.p.MGs = []
        for idx, src in enumerate(tx.sources):
            msg = MoneroTransactionSignInputRequest(
                src_entr=tmisc.translate_monero_src_entry_pb(src),
                vini=await tmisc.dump_msg(self.ct.tx.vin[idx], prefix=b"\x02"),
                vini_hmac=self.ct.tx_in_hmacs[idx],
                pseudo_out=self.ct.pseudo_outs[idx][0] if not in_memory else None,
                pseudo_out_hmac=self.ct.pseudo_outs[idx][1] if not in_memory else None,
                alpha_enc=self.ct.alphas[idx],
                spend_enc=self.ct.spend_encs[idx],
            )
            t_res = await self.trezor.tsx_sign(
                msg
            )  # type: MoneroTransactionSignInputAck
            self.handle_error(t_res)

            mg = await tmisc.parse_msg(t_res.signature, xmrtypes.MgSig())
            rv.p.MGs.append(mg)
            couts.append(t_res.cout)

        self.ct.tx.signatures = []
        self.ct.tx.rct_signatures = rv

        t_res = await self.trezor.tsx_sign(
            MoneroTransactionFinalRequest()
        )  # type: MoneroTransactionFinalAck
        self.handle_error(t_res)

        if multisig:
            cout_key = t_res.cout_key
            for ccout in couts:
                self.ct.couts.append(chacha_poly.decrypt_pack(cout_key, ccout))

        self.ct.enc_salt1, self.ct.enc_salt2 = t_res.salt, t_res.rand_mult
        self.ct.enc_keys = t_res.tx_enc_keys
        return self.ct.tx

    async def _on_all_input_set(self, ack):
        if not self._is_offloading():
            return
        if ack.rsig_data is None:
            raise ValueError("Rsig offloading requires rsig param")

        rsig_data = ack.rsig_data
        if rsig_data.mask is None:
            raise ValueError("Gamma masks not present in offloaded version")

        mask = rsig_data.mask
        if len(mask) != self._num_outputs() * 32:
            raise ValueError("Invalid number of masks")

        for i in range(len(mask) // 32):
            self.ct.rsig_gamma.append(mask[i * 32 : (i + 1) * 32])

    async def _step_set_outputs(self, idx, dst):
        self.ct.cur_output_idx = idx
        self.ct.cur_output_in_batch_idx += 1

        msg = MoneroTransactionSetOutputRequest(
            dst_entr=tmisc.translate_monero_dest_entry_pb(dst),
            dst_entr_hmac=self.ct.tx_out_entr_hmacs[idx],
        )

        # Range sig offloading to the host
        if not self._is_offloading():
            return msg

        if (
            self.ct.rsig_batches[self.ct.cur_batch_idx]
            > self.ct.cur_output_in_batch_idx
        ):
            return msg

        rsig_data = MoneroTransactionRsigData()
        batch_size = self.ct.rsig_batches[self.ct.cur_batch_idx]

        if not self._is_req_bulletproof():
            if batch_size > 1:
                raise ValueError("Borromean cannot batch outputs")

            mask = crypto.decodeint(self.ct.rsig_gamma[idx])
            C, a, R = ring_ct.prove_range_mem(dst.amount, mask)
            rsig_data.rsig = tmisc.dump_msg(R)
            self.ct.tx_out_rsigs.append(R)

        else:
            amounts = []
            masks = []
            for i in range(batch_size):
                amounts.append(
                    self.ct.tx_data.splitted_dsts[1 + idx - batch_size + i].amount
                )
                masks.append(
                    crypto.decodeint(self.ct.rsig_gamma[1 + idx - batch_size + i])
                )

            bp = await ring_ct.prove_range_bp_batch(amounts, masks)
            self.ct.tx_out_rsigs.append(bp)

            rsig_data.rsig = await tmisc.dump_msg(bp)
        msg.rsig_data = rsig_data
        return msg

    async def _on_set_outputs_ack(self, t_res):
        self.ct.tx.vout.append(await tmisc.parse_msg(t_res.tx_out, xmrtypes.TxOut()))
        self.ct.tx_out_hmacs.append(t_res.vouti_hmac)
        self.ct.tx_out_pk.append(await tmisc.parse_msg(t_res.out_pk, xmrtypes.CtKey()))
        self.ct.tx_out_ecdh.append(
            await tmisc.parse_msg(t_res.ecdh_info, xmrtypes.EcdhTuple())
        )

        rsig_buff = None
        if t_res.rsig_data:
            tsig_data = t_res.rsig_data

            if tsig_data.rsig and len(tsig_data.rsig) > 0:
                rsig_buff = tsig_data.rsig
            elif tsig_data.rsig_parts and len(tsig_data.rsig_parts) > 0:
                rsig_buff = b"".join(tsig_data.rsig_parts)

        if rsig_buff and not self._is_req_bulletproof():
            rsig = await tmisc.parse_msg(rsig_buff, xmrtypes.RangeSig())
        elif rsig_buff:
            rsig = await tmisc.parse_msg(rsig_buff, xmrtypes.Bulletproof())
        else:
            return

        if self._is_req_bulletproof():
            rsig.V = []
            batch_size = self.ct.rsig_batches[self.ct.cur_batch_idx]
            for i in range(batch_size):
                commitment = self.ct.tx_out_pk[
                    1 + self.ct.cur_output_idx - batch_size + i
                ].mask
                commitment = crypto.scalarmult(
                    crypto.decodepoint(commitment), crypto.sc_inv_eight()
                )
                rsig.V.append(crypto.encodepoint(commitment))

        self.ct.tx_out_rsigs.append(rsig)

        # Rsig verification
        try:
            if not ring_ct.ver_range(
                C=crypto.decodepoint(self.ct.tx_out_pk[-1].mask),
                rsig=rsig,
                use_bulletproof=self._is_req_bulletproof(),
            ):
                logger.warning("Rsing not valid")

        except Exception as e:
            logger.error("Exception rsig: %s" % e)
            traceback.print_exc()

        self.ct.cur_batch_idx += 1
        self.ct.cur_output_in_batch_idx = 0

    def last_transaction_data(self):
        """
        Returns last transaction data
        :return:
        """
        return self.ct

    async def get_address(self, account=0):
        """
        Get address from the device
        :return:
        """
        msg = MoneroGetAddress(
            address_n=self.address_n, network_type=self.network_type, account=account
        )
        res = await self.trezor.call(msg)
        return res

    async def get_watch_only(self):
        """
        Watch only key load
        :return:
        """
        msg = MoneroGetWatchKey(
            address_n=self.address_n, network_type=self.network_type
        )
        res = await self.trezor.get_view_key(msg)
        return res

    async def import_outputs(self, outputs):
        """
        Key images sync. Required for hot wallet be able to construct transactions.
        If the signed transaction is not relayed with the hot wallet it gets out of sync with
        key images. Thus importing is needed.

        Wallet2::import_outputs()

        :param outputs:
        :return:
        """
        ki_export_init = await key_image.generate_commitment(outputs)
        ki_export_init.address_n = self.address_n
        ki_export_init.network_type = self.network_type
        t_res = await self.trezor.key_image_sync(ki_export_init)
        self.handle_error(t_res)

        sub_res = []
        iter = await key_image.yield_key_image_data(outputs)
        batches = common.chunk(iter, 10)
        for rr in batches:  # type: list[key_image.MoneroTransferDetails]
            t_res = await self.trezor.key_image_sync(
                MoneroKeyImageSyncStepRequest(tdis=rr)
            )
            self.handle_error(t_res)

            sub_res += t_res.kis

        t_res = await self.trezor.key_image_sync(MoneroKeyImageSyncFinalRequest())
        self.handle_error(t_res)

        # Decrypting phase
        enc_key = bytes(t_res.enc_key)
        final_res = []
        for sub in sub_res:  # type: key_image.MoneroExportedKeyImage
            plain = chacha_poly.decrypt(
                enc_key, bytes(sub.iv), bytes(sub.blob), bytes(sub.tag)
            )
            ki_bin = plain[:32]

            # ki = crypto.decodepoint(ki_bin)
            # sig = [crypto.decodeint(plain[32:64]), crypto.decodeint(plain[64:])]

            final_res.append((ki_bin, (plain[32:64], plain[64:])))

        return final_res



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionData import MoneroTransactionData

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MoneroTransactionInitRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 501

    def __init__(
        self,
        version: int = None,
        address_n: List[int] = None,
        network_type: int = None,
        tsx_data: MoneroTransactionData = None,
    ) -> None:
        self.version = version
        self.address_n = address_n if address_n is not None else []
        self.network_type = network_type
        self.tsx_data = tsx_data

    @classmethod
    def get_fields(cls):
        return {
            1: ('version', p.UVarintType, 0),
            2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            3: ('network_type', p.UVarintType, 0),
            4: ('tsx_data', MoneroTransactionData, 0),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


import os
import logging

from monero_glue.hwtoken import token
from monero_glue.protocol_base.messages import MessageConverter

from trezorlib.client import TrezorClient, TrezorClientDebugLink
from trezorlib.transport import enumerate_devices, get_transport


logger = logging.getLogger(__name__)


class TrezorSession(object):
    def __init__(self, client, **kwargs):
        self.client = client

    def __enter__(self):
        self.client.transport.session_begin()
        return self.client

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.transport.session_end()


class Trezor(token.TokenLite):
    """
    Trezor proxy calls to the trezor
    """

    def __init__(
        self, path=None, debug=False, address_n=None, network_type=0, *args, **kwargs
    ):
        super().__init__()
        if path is None:
            path = os.environ.get("TREZOR_PATH", "udp:127.0.0.1:21324")

        self.debug = debug
        self.path = path
        self.msg_conv = MessageConverter(fix_bytes=True)
        self._connect()

    def _connect(self):
        self.wirelink = get_transport(self.path)
        self.client = (
            TrezorClientDebugLink(self.wirelink)
            if self.debug
            else TrezorClient(self.wirelink)
        )

        if self.debug:
            try:
                self.debuglink = self.wirelink.find_debug()
                self.client.set_debuglink(self.debuglink)
            except Exception as e:
                logger.warning(e)

    def _to_tlib(self, msg):
        return self.msg_conv.to_trezorlib(msg)

    def _from_tlib(self, msg):
        return self.msg_conv.to_phlib(msg)

    @staticmethod
    def enumerate():
        return enumerate_devices()

    def reconnect(self):
        self._connect()

    def close(self):
        self.client.close()

    def session(self):
        return TrezorSession(self.client)

    async def call(self, msg, recode=True):
        with self.session():
            return await self.call_in_session(msg, recode)

    async def call_in_session(self, msg, recode=True):
        if recode:
            msg = self._to_tlib(msg)

        res = self.client.call(msg)

        if recode:
            res = self._from_tlib(res)
        return res

    async def ping(self, message=None, **kwargs):
        with self.session():
            return self.client.ping(message if message else "monero", **kwargs)

    async def get_view_key(self, msg):
        with self.session():
            res = self.client.call(self._to_tlib(msg))
            return self._from_tlib(res)

    async def tsx_sign(self, msg):
        with self.session():
            res = self.client.call(self._to_tlib(msg))
            return self._from_tlib(res)

    async def key_image_sync(self, msg, *args, **kwargs):
        with self.session():
            res = self.client.call(self._to_tlib(msg))
            return self._from_tlib(res)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TronAccountUpdateContract import TronAccountUpdateContract
from .TronAssetIssueContract import TronAssetIssueContract
from .TronFreezeBalanceContract import TronFreezeBalanceContract
from .TronParticipateAssetIssueContract import TronParticipateAssetIssueContract
from .TronProposalApproveContract import TronProposalApproveContract
from .TronProposalCreateContract import TronProposalCreateContract
from .TronProposalDeleteContract import TronProposalDeleteContract
from .TronTransferAssetContract import TronTransferAssetContract
from .TronTransferContract import TronTransferContract
from .TronUnfreezeAssetContract import TronUnfreezeAssetContract
from .TronUnfreezeBalanceContract import TronUnfreezeBalanceContract
from .TronUpdateAssetContract import TronUpdateAssetContract
from .TronVoteWitnessContract import TronVoteWitnessContract
from .TronWithdrawBalanceContract import TronWithdrawBalanceContract
from .TronWitnessCreateContract import TronWitnessCreateContract
from .TronWitnessUpdateContract import TronWitnessUpdateContract


class TronContract(p.MessageType):

    def __init__(
        self,
        transfer_contract: TronTransferContract = None,
        transfer_asset_contract: TronTransferAssetContract = None,
        vote_witness_contract: TronVoteWitnessContract = None,
        witness_create_contract: TronWitnessCreateContract = None,
        asset_issue_contract: TronAssetIssueContract = None,
        witness_update_contract: TronWitnessUpdateContract = None,
        participate_asset_issue_contract: TronParticipateAssetIssueContract = None,
        account_update_contract: TronAccountUpdateContract = None,
        freeze_balance_contract: TronFreezeBalanceContract = None,
        unfreeze_balance_contract: TronUnfreezeBalanceContract = None,
        withdraw_balance_contract: TronWithdrawBalanceContract = None,
        unfreeze_asset_contract: TronUnfreezeAssetContract = None,
        update_asset_contract: TronUpdateAssetContract = None,
        proposal_create_contract: TronProposalCreateContract = None,
        proposal_approve_contract: TronProposalApproveContract = None,
        proposal_delete_contract: TronProposalDeleteContract = None,
    ) -> None:
        self.transfer_contract = transfer_contract
        self.transfer_asset_contract = transfer_asset_contract
        self.vote_witness_contract = vote_witness_contract
        self.witness_create_contract = witness_create_contract
        self.asset_issue_contract = asset_issue_contract
        self.witness_update_contract = witness_update_contract
        self.participate_asset_issue_contract = participate_asset_issue_contract
        self.account_update_contract = account_update_contract
        self.freeze_balance_contract = freeze_balance_contract
        self.unfreeze_balance_contract = unfreeze_balance_contract
        self.withdraw_balance_contract = withdraw_balance_contract
        self.unfreeze_asset_contract = unfreeze_asset_contract
        self.update_asset_contract = update_asset_contract
        self.proposal_create_contract = proposal_create_contract
        self.proposal_approve_contract = proposal_approve_contract
        self.proposal_delete_contract = proposal_delete_contract

    @classmethod
    def get_fields(cls):
        return {
            1: ('transfer_contract', TronTransferContract, 0),
            2: ('transfer_asset_contract', TronTransferAssetContract, 0),
            4: ('vote_witness_contract', TronVoteWitnessContract, 0),
            5: ('witness_create_contract', TronWitnessCreateContract, 0),
            6: ('asset_issue_contract', TronAssetIssueContract, 0),
            8: ('witness_update_contract', TronWitnessUpdateContract, 0),
            9: ('participate_asset_issue_contract', TronParticipateAssetIssueContract, 0),
            10: ('account_update_contract', TronAccountUpdateContract, 0),
            11: ('freeze_balance_contract', TronFreezeBalanceContract, 0),
            12: ('unfreeze_balance_contract', TronUnfreezeBalanceContract, 0),
            13: ('withdraw_balance_contract', TronWithdrawBalanceContract, 0),
            14: ('unfreeze_asset_contract', TronUnfreezeAssetContract, 0),
            15: ('update_asset_contract', TronUpdateAssetContract, 0),
            16: ('proposal_create_contract', TronProposalCreateContract, 0),
            17: ('proposal_approve_contract', TronProposalApproveContract, 0),
            18: ('proposal_delete_contract', TronProposalDeleteContract, 0),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


from monero_glue.messages import Failure


class AgentError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class StubException(Exception):
    def __init__(self, text=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    def __repr__(self):
        return "StubException(%r)" % self.text

    def __str__(self):
        return self.__repr__()


class TrezorReturnedError(AgentError):
    def __init__(self, resp=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resp = resp

    def __str__(self):
        if self.resp and isinstance(self.resp, Failure):
            return "TrezorReturnedError(Failure(code=%r, message=%r))" % (
                self.resp.code,
                self.resp.message,
            )
        else:
            return super().__str__()


class TrezorNotRunning(AgentError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import binascii
from binascii import unhexlify
import unittest

import aiounittest

from monero_glue.xmr import crypto
from monero_glue.xmr.sub.seed import SeedDerivation
from monero_glue.xmr.sub.xmr_net import NetworkTypes


class SeedTest(aiounittest.AsyncTestCase):
    """Simple seed tests"""

    def __init__(self, *args, **kwargs):
        super(SeedTest, self).__init__(*args, **kwargs)

    def recode_int(self, bts):
        return crypto.encodeint(crypto.decodeint(bts))

    def test_bip44_1(self):
        seed_wlist = "permit universe parent weapon amused modify essay borrow tobacco budget walnut lunch consider gallery ride amazing frog forget treat market chapter velvet useless topple"
        d = SeedDerivation.from_mnemonics(seed_wlist.split(" "))

        self.assertEqual(
            d.creds(network_type=NetworkTypes.MAINNET).address,
            b"42uUdtM9DyXTrGxYkyN8KW2myWJo5G71QiKyzheGHhRHY5khGGg4B7UfwzeQezzx65H9QPn2XtKqPHm7rRS1CQNpMFY3wud",
        )
        self.assertEqual(
            d.creds(network_type=NetworkTypes.TESTNET).address,
            b"9tT2891QWLdTrGxYkyN8KW2myWJo5G71QiKyzheGHhRHY5khGGg4B7UfwzeQezzx65H9QPn2XtKqPHm7rRS1CQNpMDDNrji",
        )

    def test_bip44_2(self):
        seed_wlist = "imitate lonely burden old genius trip identify wine walnut forget truly industry quarter october rug public stone rival photo more whip sample fix shrimp"
        d = SeedDerivation.from_mnemonics(seed_wlist.split(" "))

        self.assertEqual(
            d.creds(network_type=NetworkTypes.MAINNET).address,
            b"4AyUd1U6H3jSj2gbbaAyMm8wh3Xoah23x7Pm2SycZkURbHEbRoine8dE2QpJj5sEwp3YBy3z98K9t38Yzv4Fd8o69oTEnoB",
        )

    def test_slip0010_1(self):
        """
        https://github.com/satoshilabs/slips/blob/master/slip-0010.md
        :return:
        """
        seed = unhexlify(b"000102030405060708090a0b0c0d0e0f")

        d = SeedDerivation.from_master_seed(seed, slip0010=True, path="m")
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"2b4be7f19ee27bbf30c667b642d5f4aa69fd169872f8fc3059c08ebae2eb19e7"
                )
            ),
        )
        self.assertEqual(
            d.creds(network_type=NetworkTypes.MAINNET).address,
            b"44DybmHAKoojjhn4C57qqZeKQNXBGWznYhKuRHD3LvmyLCN2UxZfcTRLnCQASiyevEYg9szQqrdNeN6tXFa4xw484BSvd4r",
        )

        # m/0'
        d = SeedDerivation.from_master_seed(seed, slip0010=True, path="m/0'")
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"68e0fe46dfb67e368c75379acec591dad19df3cde26e63b93a8e704f1dade7a3"
                )
            ),
        )

        # m/0'/1'/2'/2'/1000000000'
        d = SeedDerivation.from_master_seed(
            seed, slip0010=True, path="m/0'/1'/2'/2'/1000000000'"
        )
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"8f94d394a8e8fd6b1bc2f3f49f5c47e385281d5c17e65324b0f62483e37e8793"
                )
            ),
        )

    def test_slip0010_2(self):
        """
        https://github.com/satoshilabs/slips/blob/master/slip-0010.md
        :return:
        """
        seed = unhexlify(
            b"fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542"
        )

        # m'
        d = SeedDerivation.from_master_seed(seed, slip0010=True, path="m")
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"171cb88b1b3c1db25add599712e36245d75bc65a1a5c9e18d76f9f2b1eab4012"
                )
            ),
        )

        # m/0'
        d = SeedDerivation.from_master_seed(seed, slip0010=True, path="m/0'")
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"1559eb2bbec5790b0c65d8693e4d0875b1747f4970ae8b650486ed7470845635"
                )
            ),
        )

        # m/0H/2147483647H/1H/2147483646H/2H
        d = SeedDerivation.from_master_seed(
            seed, slip0010=True, path="m/0'/2147483647'/1'/2147483646'/2'"
        )
        self.assertEqual(
            d.monero_master,
            self.recode_int(
                unhexlify(
                    b"551d333177df541ad876a60ea71f00447931c0a9da16f227c11ea080d7391b8d"
                )
            ),
        )

    def test_slip0010_3(self):
        seed_wlist = "permit universe parent weapon amused modify essay borrow tobacco budget walnut lunch consider gallery ride amazing frog forget treat market chapter velvet useless topple"
        d = SeedDerivation.from_mnemonics(seed_wlist.split(" "), slip0010=True)

        self.assertEqual(
            d.creds(network_type=NetworkTypes.MAINNET).address,
            b"497MP7bh5MkCZ1TiZZoYi84s39n5V5HpSgPvrqqUMqw6XcWDu4QhBs4VkqiPsNMhTUPAHHE6DUo9UUqpfSZfxkANPw44Fgn",
        )
        self.assertEqual(
            d.creds(network_type=NetworkTypes.TESTNET).address,
            b"9zetsNFxMirCZ1TiZZoYi84s39n5V5HpSgPvrqqUMqw6XcWDu4QhBs4VkqiPsNMhTUPAHHE6DUo9UUqpfSZfxkANPwbdsur",
        )


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class MoneroRctKey(p.MessageType):

    def __init__(
        self,
        dest: bytes = None,
        mask: bytes = None,
    ) -> None:
        self.dest = dest
        self.mask = mask

    @classmethod
    def get_fields(cls):
        return {
            1: ('dest', p.BytesType, 0),
            2: ('mask', p.BytesType, 0),
        }



# import gc
# import micropython
# import ustruct
#
# from trezor import log
#
# from apps.monero import layout
# from apps.monero.controller import misc
# from apps.monero.xmr import common, crypto, monero
# from apps.monero.xmr.enc import aescbc
# from apps.monero.xmr.serialize.int_serialize import dump_uvarint_b, load_uvarint_b

import struct as ustruct
from hashlib import sha256

from monero_glue.compat import gc, log, micropython
from monero_glue.hwtoken import misc
from monero_glue.xmr import common, crypto, monero
from monero_glue.xmr.enc import aescbc
from monero_serialize.core.int_serialize import load_uvarint_b

from .consts import *


def memcpy(dst, dst_off, src, src_off, ln):
    for i in range(ln):
        dst[dst_off + i] = src[src_off + i]
    return dst


class LiteProtocol(object):
    """
    Lite protocol
    """

    def __init__(self, **kwargs):
        self.creds = None
        self.ctx = None
        self.iface = None

        self.state = None
        self.options = None
        self.sig_mode = 0
        self.account = 0
        self.a = None
        self.A = None
        self.b = None
        self.B = None

        self.key_enc = None
        self.tx_state = None
        self.r = None
        self.R = None

        self.ctx_f = None
        self.ctx_h = None
        self.H = None
        self.c = None

        self.ctx_amount = None
        self.KV = None

        self.ctx_commitment = None
        self.C = None

        self.iv = bytearray(16)
        self.c_msg = None
        self.c_offset = 0
        self.c_p1 = 0
        self.c_p2 = 0

        self.r_msg = bytearray(2048)
        self.r_len = 0

    def _log_trace(self, x=None, collect=False):
        log.debug(
            __name__,
            "Log trace %s, ... F: %s A: %s, S: %s",
            x,
            gc.mem_free(),
            gc.mem_alloc(),
            micropython.stack_use(),
        )
        if collect:
            gc.collect()

    def assrt(self, condition, msg=None):
        """
        Asserts condition
        :param condition:
        :param msg:
        :return:
        """
        if condition:
            return
        raise ValueError("Assertion error%s" % (" : %s" % msg if msg else ""))

    async def init(self, ctx, msg):
        from apps.monero.controller import wrapper

        self.ctx = ctx
        self.creds = await wrapper.monero_get_creds(
            self.ctx, msg.address_n or (), msg.network_type
        )
        self._state_init()
        self.a = self.creds.view_key_private
        self.A = self.creds.view_key_public
        self.b = self.creds.spend_key_private
        self.B = self.creds.spend_key_public
        self.key_enc = crypto.random_bytes(
            32
        )  # WARNING: potentially dangerous for whole session.

    def _state_init(self):
        self.ctx_f = crypto.get_keccak()
        self.ctx_h = crypto.get_keccak()
        self.ctx_amount = sha256()
        self.ctx_commitment = sha256()

    def _reset_io(self):
        self.r_len = 0
        self.c_offset = 0

    def _assert_available(self, ln=32):
        if len(self.c_msg) < self.c_offset + ln:
            raise ValueError("Wrong length")

    def _fetch_view(self):
        mv = memoryview(self.c_msg)
        return mv[self.c_offset :]

    def _fetch(self, ln=32, shift=True):
        data = self.c_msg[self.c_offset : self.c_offset + ln]
        if len(data) != ln:
            raise ValueError("Wrong length")

        if shift:
            self.c_offset += ln
        return data

    def _fetch_decrypt(self, ln=32):
        data = self._fetch(ln)
        return aescbc.decrypt(self.key_enc, data, self.iv)

    def _fetch_decrypt_key(self):
        buff = self._fetch()
        if buff == b"\x00" * 32:
            return self.a
        elif buff == b"\xff" * 32:
            return self.b
        else:
            return crypto.decodeint(aescbc.decrypt(self.key_enc, buff, self.iv))

    def _fetch_u32(self):
        return (
            (self._fetch_u8() << 24)
            | (self._fetch_u8() << 16)
            | (self._fetch_u8() << 8)
            | self._fetch_u8()
        )

    def _fetch_u16(self):
        return (self._fetch_u8() << 8) | self._fetch_u8()

    def _fetch_u8(self):
        d = self._fetch(1)[0]
        return int(d) & 0xff

    def _fetch_t(self):
        t = self._fetch_u8()
        if t & 0x1f == 0x1f:
            t = (t << 8) | self._fetch_u8()
        return t

    def _fetch_l(self):
        l = self._fetch_u8()
        if l & 0x80 != 0:
            l &= 0x7f
        if l == 1:
            l = self._fetch_u8()
        elif l == 2:
            l = self._fetch_u16()
        else:
            l = -1
        return l

    def _insert(self, buff, ln=None):
        if ln is None:
            ln = len(buff)
        memcpy(self.r_msg, self.r_len, buff, 0, ln)
        self.r_len += ln

    def _insert_encrypt(self, buff, ln=None):
        enc = aescbc.encrypt(self.key_enc, buff, self.iv)
        self._insert(enc, len(enc))

    def _insert_u8(self, x):
        self.r_msg[self.r_len] = x & 0xff
        self.r_len += 1

    def _insert_u16(self, x):
        self._insert_u8(x >> 8)
        self._insert_u8(x)

    def _insert_u32(self, x):
        self._insert_u8(x >> 24)
        self._insert_u8(x >> 16)
        self._insert_u8(x >> 8)
        self._insert_u8(x)

    def _insert_t(self, t):
        if t & 0xff00:
            self._insert_u16(t)
        else:
            self._insert_u8(t)

    def _insert_tl(self, t, l):
        self._insert_t(t)
        if l < 128:
            self._insert_u8(l)
        elif l < 256:
            self._insert_u16(0x8100 | l)
        else:
            self._insert_u8(0x82)
            self._insert_u16(l)

    async def blind(self):
        AKout = self._fetch_decrypt(32)
        k = self._fetch(32)
        v = self._fetch(32)

        self.ctx_amount.update(AKout)
        self.ctx_amount.update(k)
        self.ctx_amount.update(v)

        AKout = crypto.hash_to_scalar(AKout)
        k = crypto.sc_add(crypto.decodeint(k), AKout)

        AKout = crypto.hash_to_scalar(crypto.encodeint(AKout))
        v = crypto.sc_add(crypto.decodeint(v), AKout)

        self._insert(crypto.encodeint(v))
        self._insert(crypto.encodeint(k))
        return SW_OK

    def unblind_int(self, v, k, AKout):
        AKout = crypto.hash_to_scalar(AKout)
        k = crypto.sc_sub(crypto.decodeint(k), AKout)

        AKout = crypto.hash_to_scalar(crypto.encodeint(AKout))
        v = crypto.sc_sub(crypto.decodeint(v), AKout)

        return v, k, AKout

    async def unblind(self):
        AKout = self._fetch_decrypt(32)
        k = self._fetch(32)
        v = self._fetch(32)

        v, k, AKout = self.unblind_int(v, k, AKout)

        self._insert(crypto.encodeint(v))
        self._insert(crypto.encodeint(k))
        return SW_OK

    async def put_key(self):
        sec = crypto.decodeint(self._fetch(32))
        pub = crypto.decodepoint(self._fetch(32))
        if not crypto.point_eq(pub, crypto.scalarmult_base(sec)):
            return SW_WRONG_DATA
        self.a = sec

        sec = crypto.decodeint(self._fetch(32))
        pub = crypto.decodepoint(self._fetch(32))
        if not crypto.point_eq(pub, crypto.scalarmult_base(sec)):
            return SW_WRONG_DATA
        self.b = sec
        return SW_OK

    async def get_key(self):
        if self.c_p1 == 1:
            self._insert(crypto.encodepoint(self.A))
            self._insert(crypto.encodepoint(self.B))
            self._insert(self.creds.address)

        elif self.c_p1 == 2:
            # TODO: await layout.require_confirm_watchkey(self.ctx)
            self._insert(crypto.encodeint(self.a))

        else:
            return SW_WRONG_P1P2
        return SW_OK

    async def verify_key(self):
        priv = self._fetch_decrypt_key()
        pub = crypto.decodepoint(self._fetch())
        computed_pub = crypto.identity()
        verified = 0

        if self.c_p1 == 0:
            computed_pub = crypto.scalarmult_base(priv)
        elif self.c_p1 == 1:
            pub = self.A
        elif self.c_p1 == 2:
            pub = self.B
        else:
            return SW_WRONG_P1P2

        if crypto.point_eq(computed_pub, pub):
            verified = 1

        self._insert_u32(verified)
        self._insert(self.creds.address)
        return SW_OK

    async def chacha8_prekey(self):
        abt = bytearray(65)
        memcpy(abt, 0, crypto.encodeint(self.a), 0, 32)
        memcpy(abt, 32, crypto.encodeint(self.b), 0, 32)
        abt[64] = 0x8c
        pre = crypto.keccak_hash(abt)

        # gibberish expansion to 200 bytes, different from Ledger as it uses 200B of keccak state
        for i in range(6):
            self._insert(pre)
        self._insert(pre[:8])
        return SW_OK

    async def sc_add(self):
        s1 = crypto.decodeint(self._fetch())
        s2 = crypto.decodeint(self._fetch())
        self._insert(crypto.encodeint(crypto.sc_add(s1, s2)))
        return SW_OK

    async def sc_sub(self):
        s1 = crypto.decodeint(self._fetch())
        s2 = crypto.decodeint(self._fetch())
        self._insert(crypto.encodeint(crypto.sc_sub(s1, s2)))
        return SW_OK

    async def scal_mul_key(self):
        pub = crypto.decodepoint(self._fetch())
        sec = crypto.decodeint(self._fetch_decrypt())
        self._insert(crypto.encodepoint(crypto.scalarmult(pub, sec)))
        return SW_OK

    async def scal_mul_base(self):
        sec = crypto.decodeint(self._fetch_decrypt())
        self._insert(crypto.encodepoint(crypto.scalarmult_base(sec)))
        return SW_OK

    async def generate_keypair(self):
        sec = crypto.random_scalar()
        pub = crypto.scalarmult_base(sec)
        self._insert(crypto.encodepoint(pub))
        self._insert_encrypt(crypto.encodeint(sec))
        return SW_OK

    async def secret_key_to_public_key(self):
        return await self.scal_mul_base()

    async def generate_key_derivation(self):
        pub = crypto.decodepoint(self._fetch())
        sec = self._fetch_decrypt_key()
        der = crypto.generate_key_derivation(pub, sec)
        self._insert_encrypt(crypto.encodepoint(der))
        return SW_OK

    async def derivation_to_scalar(self):
        der = crypto.decodepoint(self._fetch_decrypt())
        output_index = self._fetch_u32()
        res = crypto.derivation_to_scalar(der, output_index)
        self._insert_encrypt(crypto.encodeint(res))
        return SW_OK

    async def derive_public_key(self):
        derivation = crypto.decodepoint(self._fetch_decrypt())
        output_index = self._fetch_u32()
        pub = crypto.decodepoint(self._fetch())
        drvpub = crypto.derive_public_key(derivation, output_index, pub)
        self._insert(crypto.encodepoint(drvpub))
        return SW_OK

    async def derive_secret_key(self):
        derivation = crypto.decodepoint(self._fetch_decrypt())
        output_index = self._fetch_u32()
        sec = self._fetch_decrypt_key()
        drvsec = crypto.derive_secret_key(derivation, output_index, sec)
        self._insert_encrypt(crypto.encodeint(drvsec))
        return SW_OK

    async def generate_key_image(self):
        pub = self._fetch()
        sec = crypto.decodeint(self._fetch_decrypt())
        image = monero.generate_key_image(pub, sec)
        self._insert(crypto.encodepoint(image))
        return SW_OK

    async def derive_subaddress_public_key(self):
        pub = crypto.decodepoint(self._fetch())
        derivation = crypto.decodepoint(self._fetch_decrypt())
        output_index = self._fetch_u32()
        sub_pub = monero.derive_subaddress_public_key(pub, derivation, output_index)
        self._insert(crypto.encodepoint(sub_pub))
        return SW_OK

    def _idx_parse(self, index):
        major = ustruct.unpack("<L", index)[0]
        minor = ustruct.unpack_from("<L", index, 4)[0]
        return major, minor

    async def get_subaddress(self):
        index = self._fetch(8)
        major, minor = self._idx_parse(index)
        D, C = monero.generate_sub_address_keys(self.a, self.B, major, minor)
        self._insert(crypto.encodepoint(C))
        self._insert(crypto.encodepoint(D))
        return SW_OK

    async def get_subaddress_spend_public_key(self):
        index = self._fetch(8)
        major, minor = self._idx_parse(index)
        D = monero.get_subaddress_spend_public_key(self.a, self.B, major, minor)
        self._insert(crypto.encodepoint(D))
        return SW_OK

    async def get_subaddress_secret_key(self):
        sec = crypto.decodeint(self._fetch_decrypt())
        index = self._fetch(8)
        major, minor = self._idx_parse(index)
        sub_sec = monero.get_subaddress_secret_key(sec, major=major, minor=minor)
        self._insert_encrypt(crypto.encodeint(sub_sec))
        return SW_OK

    async def mlsag_prepare(self):
        Hi = None
        xin = None
        options = 0

        if len(self.c_msg) > 1:
            options = 1
            Hi = crypto.decodepoint(self._fetch())
            if self.options & 0x40:
                xin = crypto.decodeint(self._fetch())
            else:
                xin = crypto.decodeint(self._fetch_decrypt())

        alpha = crypto.random_scalar()
        self._insert_encrypt(crypto.encodeint(alpha))

        # ai.G
        self._insert(crypto.encodepoint(crypto.scalarmult_base(alpha)))

        if options:
            # ai * Hi
            self._insert(crypto.encodepoint(crypto.scalarmult(Hi, alpha)))
            # xin * Hi
            self._insert(crypto.encodepoint(crypto.scalarmult(Hi, xin)))
        return SW_OK

    async def mlsag_hash(self):
        if self.c_p2 == 1:
            self.ctx_h = crypto.get_keccak()
            msg = self.H
        else:
            msg = self._fetch()

        self.ctx_h.update(msg)
        if self.options & 0x80 == 0:
            c = self.ctx_h.digest()
            self.c = crypto.decodeint(c)
            self._insert(crypto.encodeint(self.c))
        return SW_OK

    async def mlsag_sign(self):
        if self.sig_mode == TRANSACTION_CREATE_FAKE:
            xin = crypto.decodeint(self._fetch())
            alpha = crypto.decodeint(self._fetch())
        elif self.sig_mode == TRANSACTION_CREATE_REAL:
            xin = crypto.decodeint(self._fetch_decrypt())
            alpha = crypto.decodeint(self._fetch_decrypt())
        else:
            raise ValueError("Invalid mode")

        ss2 = crypto.sc_mulsub(self.c, xin, alpha)
        self._insert(crypto.encodeint(ss2))
        self._insert_u32(self.sig_mode)
        return SW_OK

    async def open_tx(self):
        self.ctx_amount = sha256()
        self.account = self._fetch_u32()

        self.r = crypto.random_scalar()
        self.R = crypto.scalarmult_base(self.r)

        self._insert(crypto.encodepoint(self.R))
        self._insert_encrypt(crypto.encodeint(self.r))
        return SW_OK

    async def abort_tx(self):
        self.r = None
        self.R = None
        self.ctx_h = crypto.get_keccak()
        self.ctx_amount = sha256()
        self.ctx_commitment = sha256()
        return SW_OK

    async def set_signature_mode(self):
        self.sig_mode = TRANSACTION_CREATE_FAKE
        sig_mode = self._fetch_u8()
        if sig_mode != TRANSACTION_CREATE_FAKE and sig_mode != TRANSACTION_CREATE_REAL:
            raise ValueError("Wrong data")
        self.sig_mode = sig_mode
        self._insert_u32(sig_mode)
        return SW_OK

    async def mlsag_prehash_init(self):
        if self.c_p2 == 1:
            self.KV = self.ctx_amount.digest()
            self.ctx_amount = sha256()
            self.ctx_commitment = sha256()
            self.ctx_h = crypto.get_keccak()

        self.ctx_h.update(self.c_msg[self.c_offset :])
        if self.sig_mode == TRANSACTION_CREATE_REAL and self.c_p2 == 1:
            self._fetch_u8()
            fee = load_uvarint_b(self._fetch_view())
            # TODO: await layout.require_confirm_fee(self.ctx, fee)

        return SW_OK

    async def mlsag_prehash_update(self):
        is_subaddress = 0
        Aout = None
        Bout = None
        C = None
        v = None
        k = None
        changed = 0

        is_subaddress = bool(self._fetch_u8())
        Aout = crypto.decodepoint(self._fetch())
        Bout = crypto.decodepoint(self._fetch())
        if self.sig_mode == TRANSACTION_CREATE_REAL:
            if crypto.point_eq(Aout, self.A) and crypto.point_eq(Bout, self.B):
                changed = 1

        AKout = self._fetch_decrypt()
        C = self._fetch()
        k = self._fetch()
        v = self._fetch()

        self.ctx_h.update(k)
        self.ctx_h.update(v)

        self.ctx_amount.update(AKout)

        v, k, AKout = self.unblind_int(v, k, AKout)
        self.ctx_amount.update(crypto.encodeint(k))
        self.ctx_amount.update(crypto.encodeint(v))

        vH = crypto.scalarmult_h(v)
        kG = crypto.scalarmult_base(k)
        k = crypto.point_add(kG, vH)

        if not crypto.point_eq(k, crypto.decodepoint(C)):
            raise ValueError(SW_SECURITY_COMMITMENT_CONTROL)

        self.ctx_commitment.update(C)
        if self.options & IN_OPTION_MORE_COMMAND == 0:
            k = self.ctx_amount.digest()
            if not common.ct_equal(k, self.KV):
                raise ValueError(SW_SECURITY_AMOUNT_CHAIN_CONTROL)

            self.C = self.ctx_commitment.digest()
            self.ctx_commitment = sha256()

        if self.sig_mode == TRANSACTION_CREATE_REAL:
            if not changed:
                await self._req_dst(Aout, Bout, v, is_subaddress)

        return SW_OK

    async def _req_dst(self, Aout, Bout, amount, is_sub=False):
        addr = misc.StdObj(view_public_key=Aout, spend_public_key=Bout)
        out = misc.StdObj(addr=addr, amount=amount, is_subaddress=is_sub)
        await self.iface.confirm_out(out, False, self.creds, self.ctx)

    async def mlsag_prehash_finalize(self):
        H = None
        proof = None
        message = None

        if self.options & IN_OPTION_MORE_COMMAND:
            H = self._fetch()
            self.ctx_h.update(H)
            self.ctx_commitment.update(H)
        else:
            H = self.ctx_commitment.digest()
            if not common.ct_equal(H, self.C):
                raise ValueError(SW_SECURITY_COMMITMENT_CHAIN_CONTROL)

            H = self.ctx_h.digest()
            message = self._fetch()
            proof = self._fetch()

            self.ctx_h = crypto.get_keccak()
            self.ctx_h.update(message)
            self.ctx_h.update(H)
            self.ctx_h.update(proof)
            self.H = self.ctx_h.digest()
            self._insert(self.H)
        return SW_OK

    async def stealth(self):
        pub = crypto.decodepoint(self._fetch())
        sec = self._fetch_decrypt_key()
        pay_id = self._fetch(8)

        drv = crypto.generate_key_derivation(pub, sec)
        drv += b"\x8d"
        sec = crypto.keccak_hash(drv)
        for i in range(8):
            pay_id[i] ^= sec[i]

        self._insert(pay_id)
        return SW_OK

    async def dispatch(self, ctx, ins, p1, p2, input):
        """
        Main message dispatcher
        :param ctx:
        :param ins:
        :param p1:
        :param p2:
        :param input:
        :return:
        """
        self.c_p1 = p1
        self.c_p2 = p2
        self.c_msg = input
        self.c_offset = 0
        self.r_len = 0

        from apps.monero.controller import iface

        self.ctx = ctx
        self.iface = iface.get_iface(ctx)

        sw = 0x6F01
        try:
            log.debug(__name__, "Ins: %s, %s %s", ins, p1, p2)
            sw = await self._sub_dispatch(ins, p1)
        except Exception as e:
            log.error(__name__, "Exception dispatching: %s", e)
            raise

        gc.collect()
        return sw, self.r_msg[: self.r_len]

    async def _sub_dispatch(self, ins, p1):
        sw = 0x6F01
        self.options = self._fetch_u8()

        if ins == INS_NONE:
            pass
        elif ins == INS_OPEN_TX:
            sw = await self.open_tx()
        elif ins == INS_CLOSE_TX:
            sw = SW_OK
        elif ins == INS_SET_SIGNATURE_MODE:
            sw = await self.set_signature_mode()
        elif ins == INS_STEALTH:
            sw = await self.stealth()

        elif ins == INS_PUT_KEY:
            sw = await self.put_key()
        elif ins == INS_GET_KEY:
            sw = await self.get_key()
        elif ins == INS_MANAGE_SEEDWORDS:
            sw = SW_OK  # TODO:

        elif ins == INS_VERIFY_KEY:
            sw = await self.verify_key()
        elif ins == INS_GET_CHACHA8_PREKEY:
            sw = await self.chacha8_prekey()
        elif ins == INS_SECRET_KEY_TO_PUBLIC_KEY:
            sw = await self.secret_key_to_public_key()
        elif ins == INS_GEN_KEY_DERIVATION:
            sw = await self.generate_key_derivation()
        elif ins == INS_DERIVATION_TO_SCALAR:
            sw = await self.derivation_to_scalar()
        elif ins == INS_DERIVE_PUBLIC_KEY:
            sw = await self.derive_public_key()
        elif ins == INS_DERIVE_SECRET_KEY:
            sw = await self.derive_secret_key()
        elif ins == INS_GEN_KEY_IMAGE:
            sw = await self.generate_key_image()
        elif ins == INS_SECRET_KEY_ADD:
            sw = await self.sc_add()
        elif ins == INS_SECRET_KEY_SUB:
            sw = await self.sc_sub()
        elif ins == INS_GENERATE_KEYPAIR:
            sw = await self.generate_keypair()
        elif ins == INS_SECRET_SCAL_MUL_KEY:
            sw = await self.scal_mul_key()
        elif ins == INS_SECRET_SCAL_MUL_BASE:
            sw = await self.scal_mul_base()

        elif ins == INS_DERIVE_SUBADDRESS_PUBLIC_KEY:
            sw = await self.derive_subaddress_public_key()
        elif ins == INS_GET_SUBADDRESS:
            sw = await self.get_subaddress()
        elif ins == INS_GET_SUBADDRESS_SPEND_PUBLIC_KEY:
            sw = await self.get_subaddress_spend_public_key()
        elif ins == INS_GET_SUBADDRESS_SECRET_KEY:
            sw = await self.get_subaddress_secret_key()

        elif ins == INS_BLIND:
            sw = await self.blind()
        elif ins == INS_UNBLIND:
            sw = await self.unblind()

        elif ins == INS_VALIDATE:
            if p1 == 1:
                sw = await self.mlsag_prehash_init()
            elif p1 == 2:
                sw = await self.mlsag_prehash_update()
            elif p1 == 3:
                sw = await self.mlsag_prehash_finalize()
            else:
                raise ValueError(SW_WRONG_P1P2)

        elif ins == INS_MLSAG:
            if p1 == 1:
                sw = await self.mlsag_prepare()
            elif p1 == 2:
                sw = await self.mlsag_hash()
            elif p1 == 3:
                sw = await self.mlsag_sign()
            else:
                raise ValueError(SW_WRONG_P1P2)
        return sw



class TsxSignStateHolder(object):
    """
    Simple transaction signer state holder.
    Externalized state uses smaller amount of memory compared to storing the builder instance in the state.
    Moreover the state contains stripped down attributes, i.e., instead of heavy hashers only sha3 context
    is preserved and hashers are re-initialized on the next protocol step.
    """

    def __init__(self, **kwargs):
        self.creds = None
        self.key_master = None
        self.key_hmac = None
        self.key_enc = None

        self.r = None  # txkey
        self.r_pub = None
        self.state = None

        self.multi_sig = False
        self.need_additional_txkeys = False
        self.use_bulletproof = False
        self.use_rct = True
        self.use_simple_rct = False
        self.input_count = 0
        self.output_count = 0
        self.output_change = None
        self.mixin = 0
        self.fee = 0
        self.account_idx = 0

        self.additional_tx_private_keys = []
        self.additional_tx_public_keys = []
        self.inp_idx = -1
        self.out_idx = -1
        self.summary_inputs_money = 0
        self.summary_outs_money = 0
        self.input_secrets = []
        self.input_alphas = []
        self.input_pseudo_outs = []
        self.output_sk = []
        self.output_pk = []
        self.output_amounts = []
        self.output_masks = []
        self.rsig_type = 0
        self.rsig_grp = []
        self.rsig_offload = 0
        self.sumout = None
        self.sumpouts_alphas = None
        self.subaddresses = {}
        self.tx = None
        self.source_permutation = []  # sorted by key images
        self.tx_prefix_hasher = None
        self.tx_prefix_hash = None
        self.full_message_hasher = None
        self.full_message = None
        self.exp_tx_prefix_hash = None



import os
import glob
import subprocess
import sys
import logging

from setuptools import setup, Command, find_packages
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from distutils.errors import DistutilsError


logger = logging.getLogger(__name__)


version = "1.6.4"

# Please update tox.ini when modifying dependency version requirements
install_requires = [
    "monero-serialize>=2.0.7",
    "pycryptodome",
    "py-cryptonight>=0.1.2",
    "chacha20poly1305",
]

dev_extras = [
    "nose",
    "pep8",
    "tox",
    "aiounittest",
    "requests",
    "pympler",
    "pypandoc",
    "pandoc",
    "pip",
    "pycparser",
    "ctypeslib2",
    "cryptography",  # chacha20poly1305
    "protobuf==3.4.0",
]

poc_extras = [
    "ecdsa",
    "asyncio",
    "requests",
    "cmd2>=0.6.9",
    "shellescape",
    "coloredlogs",
    "blessed>=1.14.1",
    "flask>=0.12",
    "flask-socketio",
    "eventlet",
    "gevent",
    "sarge>=0.1.5",
]

# trezor-crypto backend, secure and fast
tcry_extras = [
    "py_trezor_crypto_ph4==0.1.1",
]

docs_extras = [
    "Sphinx>=1.0",  # autodoc_member_order = 'bysource', autodoc_default_flags
    "sphinx_rtd_theme",
    "sphinxcontrib-programoutput",
]

# trezor dependencies, trezorlib
trezorlib_git = 'https://github.com/trezor/python-trezor.git'
trezorlib_commit = '20e0acbc98b95936d8ac63f7f653624aae29d1c0'
trezor_extras = [
    "trezor",
]


CWD = os.path.dirname(os.path.realpath(__file__))
TREZOR_COMMON = os.path.join(CWD, 'vendor', 'trezor-common')


class InstallTrezorlib(Command):
    description = 'installs trezorlib with the exact needed version'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import shutil
        import tempfile
        import subprocess
        tmpd = tempfile.mkdtemp()

        try:
            subprocess.call(["git", "clone", trezorlib_git, "."], cwd=tmpd)
            subprocess.call(["git", "checkout", trezorlib_commit], cwd=tmpd)
            subprocess.call(["git", "submodule", "init"], cwd=tmpd)
            subprocess.call(["git", "submodule", "update", "--recursive"], cwd=tmpd)

            # Not nice, but due to virtualenv we dont know which pip is being used
            try:
                from pip._internal import main as pip_main
                pip_main(["install", "-U", tmpd])

            except Exception as e:
                logger.warning('Programatical pip installation failed: %s' % e)

                pip_path = os.getenv('PIP_PATH', None)
                pip = os.path.join(pip_path, 'pip') if pip_path else 'pip'
                subprocess.call([pip, "install", "-U", tmpd])

        except Exception as e:
            msg = 'Error installing "trezor" package. Try the manual way: ' \
                  'git clone %s /tmp/trezorlib && cd /tmp/trezorlib && git checkout %s && ' \
                  'git submodule init && git submodule update --recursive && ' \
                  'cd - && pip install -U /tmp/trezorlib' % (trezorlib_git, trezorlib_commit)
            raise DistutilsError(msg) from e
        finally:
            shutil.rmtree(tmpd)


class PrebuildCommand(Command):
    description = 'update vendored files (coins.json, protobuf messages)'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # check for existence of the submodule directory
        common_defs = os.path.join(TREZOR_COMMON, 'defs')
        if not os.path.exists(common_defs):
            raise DistutilsError('trezor-common submodule seems to be missing.\n' +
                                 'Use "git submodule update --init" to retrieve it.')

        # regenerate messages
        try:
            proto_srcs = glob.glob(os.path.join(TREZOR_COMMON, "protob", "*.proto"))
            subprocess.check_call([
                sys.executable,
                os.path.join(TREZOR_COMMON, "protob", "pb2py"),
                "-o", os.path.join(CWD, "monero_glue", "messages"),
                "-P", "..protobuf",
            ] + proto_srcs)
        except Exception as e:
            raise DistutilsError("Generating protobuf failed. Make sure you have 'protoc' in your PATH.") from e


def _patch_prebuild(cls):
    """Patch a setuptools command to depend on `prebuild`"""
    orig_run = cls.run

    def new_run(self):
        self.run_command('prebuild')
        orig_run(self)

    cls.run = new_run


_patch_prebuild(build_py)
_patch_prebuild(develop)


try:
    import pypandoc

    long_description = pypandoc.convert("README.md", "rst")
    long_description = long_description.replace("\r", "")

except (IOError, ImportError):
    import io

    with io.open("README.md", encoding="utf-8") as f:
        long_description = f.read()

setup(
    name="monero_agent",
    version=version,
    description="Monero Agent",
    long_description=long_description,
    url="https://github.com/ph4r05/monero-agent",
    author="Dusan Klinec",
    author_email="dusan.klinec@gmail.com",
    license="CLOSED",
    # license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Security",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require={
        "dev": dev_extras,
        "poc": poc_extras,
        "docs": docs_extras,
        "tcry": tcry_extras,
        "trezor": trezor_extras,
    },
    cmdclass={
        'prebuild': PrebuildCommand,
        'install_trezorlib': InstallTrezorlib,
    },
    entry_points={
        'console_scripts': [
            'monero-seed = monero_poc.seed:main',
            'monero-agent = monero_poc.agent:main',
            'monero-trezor-poc = monero_poc.trezor:main',
        ],
    }

)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import ctypes as ct
import hmac

from Crypto.Protocol.KDF import PBKDF2
from monero_glue.xmr.core.ec_base import *
from trezor_crypto import trezor_cfunc as tcryr

# from monero_glue.misc.devel.call_tracker import CallTracker
# tcry = CallTracker(tcryr, print_on_end=True)
tcry = tcryr

ED25519_ORD = b"\xed\xd3\xf5\x5c\x1a\x63\x12\x58\xd6\x9c\xf7\xa2\xde\xf9\xde\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10"
ED25519_ORD_BN = None


# Load & init the library
tcry.open_lib()


def new_point():
    return tcry.new_ge25519()


def new_scalar():
    return tcry.init256_modm_r(0)


def random_bytes(by):
    """
    Generates X random bytes, returns byte-string
    :param by:
    :return:
    """
    return tcry.random_buffer_r(by)


class KeccakWrapper(object):
    """
    Simple Keccak hasher wrapper. OOP interface to xmr_hasher_*
    """

    digest_size = 32
    block_size = 136  # 1088 bits

    def __init__(self, h=None):
        self.h = tcry.xmr_hasher_init_r() if h is None else h

    def __repr__(self):
        return "<KeccakHash: %s>" % self.h

    def copy(self):
        h_copy = tcry.xmr_hasher_copy_r(self.h)
        return KeccakWrapper(h_copy)

    def update(self, s):
        tcry.xmr_hasher_update(self.h, bytes(s))

    def digest(self):
        r = tcry.xmr_hasher_final_r(self.h)
        self.h = None
        return r

    def hexdigest(self):
        return self.digest().encode("hex")


def get_keccak():
    """
    Simple keccak 256
    :return:
    """
    return KeccakWrapper()


def keccak_hash(inp):
    """
    Hashesh input in one call
    :return:
    """
    return tcry.xmr_fast_hash_r(bytes(inp))


def keccak_hash_into(r, inp):
    """
    Hashesh input in one call
    :return:point_add
    """
    bf = tcry.KEY_BUFF.from_buffer(r)
    tcry.cl().xmr_fast_hash(bf, bytes(inp), len(inp))
    return r


def keccak_2hash(inp):
    """
    Keccak double hashing
    :param inp:
    :return:
    """
    return keccak_hash(keccak_hash(inp))


def get_hmac(key, msg=None):
    """
    Returns HMAC object (uses Keccak256)
    :param key:
    :param msg:
    :return:
    """
    return hmac.new(key, msg=msg, digestmod=get_keccak)


def compute_hmac(key, msg=None):
    """
    Computes and returns HMAC of the msg using Keccak256
    :param key:
    :param msg:
    :return:
    """
    h = hmac.new(key, msg=msg, digestmod=get_keccak)
    return h.digest()


def pbkdf2(inp, salt, length=32, count=1000, prf=None):
    """
    PBKDF2 with default PRF as HMAC-KECCAK-256
    :param inp:
    :param salt:
    :param length:
    :param count:
    :param prf:
    :return:
    """

    if prf is None:
        prf = lambda p, s: hmac.new(p, msg=s, digestmod=get_keccak).digest()
    return PBKDF2(inp, salt, length, count, prf)


#
# EC
#


def _offset(x, offset=0):
    if offset == 0:
        return x
    return memoryview(x)[offset:]


def decodepoint(x, offset=0):
    return tcry.ge25519_unpack_vartime_r(tcry.KEY_BUFF(*_offset(x, offset)[:32]))


def decodepoint_into(r, x, offset=0):
    tcry.ge25519_unpack_vartime(r, tcry.KEY_BUFF(*_offset(x, offset)[:32]))
    return r


def encodepoint(pt):
    return tcry.ge25519_pack_r(pt)


def encodepoint_into(b, pt, offset=0):
    bf = tcry.KEY_BUFF.from_buffer(_offset(b, offset))
    tcry.ge25519_pack(bf, pt)
    return b


def decodeint(x, offset=0):
    return tcry.expand256_modm_r(tcry.KEY_BUFF(*_offset(x, offset)))


def decodeint_noreduce(x, offset=0):
    return tcry.expand_raw256_modm_r(tcry.KEY_BUFF(*_offset(x, offset)))


def decodeint_into(r, x, offset=0):
    bb = _offset(x, offset)[:32]
    tcry.expand256_modm(r, tcry.KEY_BUFF(*bb), 32)
    return r


def decodeint_into_noreduce(r, x, offset=0):
    tcry.expand_raw256_modm(r, tcry.KEY_BUFF(*_offset(x, offset)))
    return r


def encodeint(x):
    return tcry.contract256_modm_r(x)


def encodeint_into(b, x, offset=0):
    bf = tcry.KEY_BUFF.from_buffer(_offset(b, offset))
    tcry.contract256_modm(bf, x)
    return b


#
# Zmod(2^255 - 19) operations, fe (field element)
# Not constant time! PoC only.
#


def fe_1():
    return tcry.curve25519_set_r(1)


def fe_mod(a):
    return tcry.curve25519_reduce_r(a)


def fe_add(a, b):
    return tcry.curve25519_add_r(a, b)


def fe_sub(a, b):
    return tcry.curve25519_sub_reduce_r(a, b)


def fe_sq(a):
    return tcry.curve25519_square_r(a)


def fe_mul(a, b):
    return tcry.curve25519_mul_r(a, b)


def fe_expmod(b, e):
    raise ValueError("Not implemented")


def fe_divpowm1(u, v):
    raise ValueError("Not implemented")


def fe_isnegative(x):
    return tcry.curve25519_isnegative(x)


def fe_isnonzero(x):
    return tcry.curve25519_isnonzero(x)


#
# Zmod(order), scalar values field
#


def sc_0():
    """
    Sets 0 to the scalar value Zmod(m)
    :return:
    """
    return tcry.init256_modm_r(0)


def sc_0_into(r):
    """
    Sets 0 to the scalar value Zmod(m)
    :return:
    """
    tcry.init256_modm(r, 0)
    return r


def sc_init(x):
    """
    Sets x to the scalar value Zmod(m)
    :return:
    """
    if x >= (1 << 64):
        raise ValueError("Initialization works up to 64-bit only")
    return tcry.init256_modm_r(x)


def sc_init_into(r, x):
    """
    Sets x to the scalar value Zmod(m)
    :return:
    """
    if x >= (1 << 64):
        raise ValueError("Initialization works up to 64-bit only")
    tcry.init256_modm(r, x)
    return r


def sc_get64(x):
    """
    Returns 64bit value from the sc
    :param x:
    :return:
    """
    return tcry.get256_modm_r(x)


def sc_check(key):
    """
    sc_check is not relevant for long-integer scalar representation.

    :param key:
    :return:
    """
    return not tcry.check256_modm(key)


def check_sc(key):
    """
    throws exception on invalid key
    :param key:
    :return:
    """
    if sc_check(key) != 0:
        raise ValueError("Invalid scalar value")


def sc_reduce32(data):
    return tcry.barrett_reduce256_modm_r(sc_0(), data)


def sc_add(aa, bb):
    return tcry.add256_modm_r(aa, bb)


def sc_add_into(r, aa, bb):
    tcry.add256_modm(r, aa, bb)
    return r


def sc_sub(aa, bb):
    return tcry.sub256_modm_r(aa, bb)


def sc_sub_into(r, aa, bb):
    tcry.sub256_modm(r, aa, bb)
    return r


def sc_mul(aa, bb):
    return tcry.mul256_modm_r(aa, bb)


def sc_mul_into(r, aa, bb):
    tcry.mul256_modm(r, aa, bb)
    return r


def sc_isnonzero(c):
    return not tcry.iszero256_modm(c)


def sc_eq(a, b):
    return tcry.eq256_modm(a, b)


def sc_mulsub(aa, bb, cc):
    """
    (cc - aa * bb) % l
    """
    return tcry.mulsub256_modm_r(aa, bb, cc)


def sc_mulsub_into(r, aa, bb, cc):
    """
    (cc - aa * bb) % l
    """
    tcry.mulsub256_modm(r, aa, bb, cc)
    return r


def sc_muladd(aa, bb, cc):
    """
    (cc + aa * bb) % l
    """
    return tcry.muladd256_modm_r(aa, bb, cc)


def sc_muladd_into(r, aa, bb, cc):
    """
    (cc + aa * bb) % l
    """
    tcry.muladd256_modm(r, aa, bb, cc)
    return r


def sc_inv(x):
    return sc_inv_into(new_scalar(), x)


def sc_inv_into(r, x):
    """
    Modular inversion mod curve order L
    :param r:
    :param x:
    :return:
    """
    global ED25519_ORD_BN
    if ED25519_ORD_BN is None:
        bf_p = tcry.KEY_BUFF.from_buffer(bytearray(ED25519_ORD))
        ED25519_ORD_BN = tcry.bn_zero_r()
        tcry.bn_read_le(bf_p, ED25519_ORD_BN)

    rr = bytearray(32)
    xx = bytearray(32)
    encodeint_into(xx, x)
    bf_x = tcry.KEY_BUFF.from_buffer(xx)
    bf_r = tcry.KEY_BUFF.from_buffer(rr)

    bn_x = tcry.bn_zero_r()
    tcry.bn_read_le(bf_x, bn_x)
    tcry.bn_inverse(bn_x, ED25519_ORD_BN)
    tcry.bn_write_le(bn_x, bf_r)
    return decodeint_into_noreduce(r, rr)


def random_scalar():
    return tcry.xmr_random_scalar_r()


def random_scalar_into(r):
    tcry.xmr_random_scalar(r)
    return r


#
# GE - ed25519 group
#

INV_EIGHT = b"\x79\x2f\xdc\xe2\x29\xe5\x06\x61\xd0\xda\x1c\x7d\xb3\x9d\xd3\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06"
INV_EIGHT_SC = decodeint(INV_EIGHT)


def check_ed25519point(x):
    if tcry.ge25519_check(x) != 1:
        raise ValueError("P is not on ed25519 curve")


def scalarmult_base(a):
    return tcry.ge25519_scalarmult_base_wrapper_r(a)


def scalarmult_base_into(r, a):
    tcry.ge25519_scalarmult_base_wrapper(r, a)
    return r


def scalarmult(P, e):
    return tcry.ge25519_scalarmult_r(P, e)


def scalarmult_into(r, P, e):
    tcry.ge25519_scalarmult(r, P, e)
    return r


def point_add(P, Q):
    return tcry.ge25519_add_r(P, Q, 0)


def point_add_into(r, P, Q):
    tcry.ge25519_add(r, P, Q, 0)
    return r


def point_sub(P, Q):
    return tcry.ge25519_add_r(P, Q, 1)


def point_sub_into(r, P, Q):
    tcry.ge25519_add(r, P, Q, 1)
    return r


def point_eq(P, Q):
    return tcry.ge25519_eq(P, Q)


def point_double(P):
    return tcry.ge25519_double_r(P)


def point_double_into(r, P):
    tcry.ge25519_double(r, P)
    return r


def point_norm(P):
    """
    Normalizes point after multiplication
    Extended edwards coordinates (X,Y,Z,T)
    :param P:
    :return:
    """
    return tcry.ge25519_norm_r(P)


def point_mul8(P):
    return tcry.ge25519_mul8_r(P)


def point_mulinv8(P):
    return scalarmult(P, INV_EIGHT_SC)


def point_mul8_into(r, P):
    tcry.ge25519_mul8(r, P)
    return r


def sc_inv_eight():
    return INV_EIGHT_SC


def ge_double_scalarmult_base_vartime(a, A, b):
    """
    void ge25519_double_scalarmult_vartime(ge25519 *r, const ge25519 *p1, const bignum256modm s1, const bignum256modm s2);
    r = a * A + b * B

    :param a:
    :param A:
    :param b:
    :return:
    """
    R = tcry.ge25519_double_scalarmult_vartime_r(A, a, b)
    tcry.ge25519_norm(R, R)
    return R


def ge_double_scalarmult_base_vartime2(a, A, b, B):
    """
    void ge25519_double_scalarmult_vartime2(ge25519 *r, const ge25519 *p1, const bignum256modm s1, const ge25519 *p2, const bignum256modm s2);
    r = a * A + b * B

    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    R = tcry.ge25519_double_scalarmult_vartime2_r(A, a, B, b)
    tcry.ge25519_norm(R, R)
    return R


def ge_double_scalarmult_precomp_vartime(a, A, b, Bi):
    """
    void ge_double_scalarmult_precomp_vartime(ge_p2 *r, const unsigned char *a, const ge_p3 *A, const unsigned char *b, const ge_dsmp Bi)
    :return:
    """
    return ge_double_scalarmult_precomp_vartime2(a, A, b, Bi)


def ge_double_scalarmult_precomp_vartime2(a, Ai, b, Bi):
    """
    void ge_double_scalarmult_precomp_vartime2(ge_p2 *r, const unsigned char *a, const ge_dsmp Ai, const unsigned char *b, const ge_dsmp Bi)
    :param a:
    :param Ai:
    :param b:
    :param Bi:
    :return:
    """
    return tcry.xmr_add_keys3_r(a, Ai, b, Bi)


def identity(byte_enc=False):
    """
    Identity point
    :return:
    """
    idd = tcry.ge25519_set_neutral_r()
    return idd if not byte_enc else encodepoint(idd)


def identity_into(r):
    """
    Identity point
    :return:
    """
    tcry.ge25519_set_neutral(r)
    return r


def ge_frombytes_vartime_check(point):
    """
    https://www.imperialviolet.org/2013/12/25/elligator.html
    http://elligator.cr.yp.to/
    http://elligator.cr.yp.to/elligator-20130828.pdf
    :param key:
    :return:
    """
    return 0 if tcry.ge25519_check(point) == 1 else -1


def ge_frombytes_vartime(point):
    """
    https://www.imperialviolet.org/2013/12/25/elligator.html

    :param key:
    :return:
    """
    ge_frombytes_vartime_check(point)
    return point


def precomp(point):
    """
    Precomputation placeholder
    :param point:
    :return:
    """
    return point


def ge_dsm_precomp(point):
    """
    void ge_dsm_precomp(ge_dsmp r, const ge_p3 *s)
    :param point:
    :return:
    """
    return point


#
# Monero specific
#


def cn_fast_hash(buff):
    """
    Keccak 256, original one (before changes made in SHA3 standard)
    :param buff:
    :return:
    """
    return keccak_hash(buff)


def hash_to_scalar(data, length=None):
    """
    H_s(P)
    :param data:
    :param length:
    :return:
    """
    dt = data[:length] if length else data
    return tcry.xmr_hash_to_scalar_r(bytes(dt))


def hash_to_scalar_into(r, data, length=None):
    """
    H_s(P)
    :param data:
    :param length:
    :return:
    """
    dt = data[:length] if length else data
    tcry.xmr_hash_to_scalar(r, bytes(dt))
    return r


def hash_to_point(buf):
    """
    H_p(buf)

    https://github.com/monero-project/research-lab/blob/master/whitepaper/ge_fromfe_writeup/ge_fromfe.pdf
    http://archive.is/yfINb
    :param key:
    :return:
    """
    return tcry.xmr_hash_to_ec_r(buf)


def hash_to_point_into(r, buf):
    """
    H_p(buf)

    https://github.com/monero-project/research-lab/blob/master/whitepaper/ge_fromfe_writeup/ge_fromfe.pdf
    http://archive.is/yfINb
    :param key:
    :return:
    """
    bf = tcry.KEY_BUFF.from_buffer(buf)
    tcry.xmr_hash_to_ec(r, bf)
    return r


#
# XMR
#


def xmr_H():
    """
    Returns point H
    8b655970153799af2aeadc9ff1add0ea6c7251d54154cfa92c173a0dd39c1f94
    :return:
    """
    return tcry.ge25519_set_xmr_h_r()


def scalarmult_h(i):
    return scalarmult(xmr_H(), sc_init(i) if isinstance(i, int) else i)


def add_keys2(a, b, B):
    """
    aG + bB, G is basepoint
    :param a:
    :param b:
    :param B:
    :return:
    """
    return tcry.xmr_add_keys2_vartime_r(a, b, B)


def add_keys2_into(r, a, b, B):
    """
    aG + bB, G is basepoint
    :param r:
    :param a:
    :param b:
    :param B:
    :return:
    """
    tcry.xmr_add_keys2_vartime(r, a, b, B)
    return r


def add_keys3(a, A, b, B):
    """
    aA + bB
    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    return tcry.xmr_add_keys3_vartime_r(a, A, b, B)


def add_keys3_into(r, a, A, b, B):
    """
    aA + bB
    :param r:
    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    tcry.xmr_add_keys3_vartime(r, a, A, b, B)
    return r


def gen_c(a, amount):
    """
    Generates Pedersen commitment
    C = aG + bH

    :param a:
    :param amount:
    :return:
    """
    return tcry.xmr_gen_c_r(a, amount)


def generate_key_derivation(key1, key2):
    """
    Key derivation: 8*(key2*key1)

    :param key1: public key of receiver Bob (see page 7)
    :param key2: Alice's private
    :return:
    """
    if sc_check(key2) != 0:
        # checks that the secret key is uniform enough...
        raise ValueError("error in sc_check in keyder")
    if ge_frombytes_vartime_check(key1) != 0:
        raise ValueError("didn't pass curve checks in keyder")

    return tcry.xmr_generate_key_derivation_r(key1, key2)


def derivation_to_scalar(derivation, output_index):
    """
    H_s(derivation || varint(output_index))
    :param derivation:
    :param output_index:
    :return:
    """
    check_ed25519point(derivation)
    return tcry.xmr_derivation_to_scalar_r(derivation, output_index)


def derive_public_key(derivation, output_index, base):
    """
    H_s(derivation || varint(output_index))G + base

    :param derivation:
    :param output_index:
    :param base:
    :return:
    """
    if ge_frombytes_vartime_check(base) != 0:  # check some conditions on the point
        raise ValueError("derive pub key bad point")
    check_ed25519point(base)

    return tcry.xmr_derive_public_key_r(derivation, output_index, base)


def derive_secret_key(derivation, output_index, base):
    """
    base + H_s(derivation || varint(output_index))
    :param derivation:
    :param output_index:
    :param base:
    :return:
    """
    if sc_check(base) != 0:
        raise ValueError("cs_check in derive_secret_key")
    return tcry.xmr_derive_private_key_r(derivation, output_index, base)


def get_subaddress_secret_key(secret_key, major=0, minor=0):
    """
    Builds subaddress secret key from the subaddress index
    Hs(SubAddr || a || index_major || index_minor)

    :param secret_key:
    :param index:
    :param major:
    :param minor:
    :return:
    """
    return tcry.xmr_get_subaddress_secret_key_r(major, minor, secret_key)


def prove_range(amount, last_mask=None):
    """
    Range proof provided by the backend. Implemented in C for speed.

    :param amount:
    :param last_mask:
    :return:
    """
    C, a, R = tcry.gen_range_proof(amount, last_mask)

    # Trezor micropython extmod returns byte-serialized/flattened rsig
    nrsig = b""
    for i in range(len(R.asig.s0)):
        nrsig += bytes(R.asig.s0[i])
    for i in range(len(R.asig.s1)):
        nrsig += bytes(R.asig.s1[i])
    nrsig += bytes(R.asig.ee)
    for i in range(len(R.Ci)):
        nrsig += bytes(R.Ci[i])
    return C, a, nrsig

    # # Rewrap to serializable structures
    # nrsig = xmrtypes.RangeSig()
    # nrsig.asig = xmrtypes.BoroSig()
    # nrsig.asig.ee = bytes(R.asig.ee)
    # nrsig.Ci = list(R.Ci)
    # nrsig.asig.s0 = list(R.asig.s0)
    # nrsig.asig.s1 = list(R.asig.s1)
    # del R
    #
    # for i in range(64):
    #     nrsig.Ci[i] = bytes(nrsig.Ci[i])
    #     nrsig.asig.s0[i] = bytes(nrsig.asig.s0[i])
    #     nrsig.asig.s1[i] = bytes(nrsig.asig.s1[i])
    #
    # return C, a, nrsig


#
# Backend config
#


class TcryECBackend(ECBackendBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def has_rangeproof_borromean(self):
        return True

    def has_rangeproof_bulletproof(self):
        return True

    def has_crypto_into_functions(self):
        return True

    def is_fast(self):
        return True


BACKEND_OBJ = None


def get_backend():
    global BACKEND_OBJ
    if BACKEND_OBJ is None:
        BACKEND_OBJ = TcryECBackend()
    return BACKEND_OBJ



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionDestinationEntry import MoneroTransactionDestinationEntry

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class MoneroTransactionRsigData(p.MessageType):

    def __init__(
        self,
        version: int = None,
        rsig_type: int = None,
        offload_type: int = None,
        grouping: List[int] = None,
        step: int = None,
        operation: int = None,
        seed: bytes = None,
        mask: bytes = None,
        amount: bytes = None,
        rsig: bytes = None,
        rsig_parts: List[bytes] = None,
        outputs: List[MoneroTransactionDestinationEntry] = None,
    ) -> None:
        self.version = version
        self.rsig_type = rsig_type
        self.offload_type = offload_type
        self.grouping = grouping if grouping is not None else []
        self.step = step
        self.operation = operation
        self.seed = seed
        self.mask = mask
        self.amount = amount
        self.rsig = rsig
        self.rsig_parts = rsig_parts if rsig_parts is not None else []
        self.outputs = outputs if outputs is not None else []

    @classmethod
    def get_fields(cls):
        return {
            1: ('version', p.UVarintType, 0),
            2: ('rsig_type', p.UVarintType, 0),
            3: ('offload_type', p.UVarintType, 0),
            4: ('grouping', p.UVarintType, p.FLAG_REPEATED),
            5: ('step', p.UVarintType, 0),
            6: ('operation', p.UVarintType, 0),
            7: ('seed', p.BytesType, 0),
            8: ('mask', p.BytesType, 0),
            9: ('amount', p.BytesType, 0),
            10: ('rsig', p.BytesType, 0),
            11: ('rsig_parts', p.BytesType, p.FLAG_REPEATED),
            12: ('outputs', MoneroTransactionDestinationEntry, p.FLAG_REPEATED),
        }
